#!/usr/bin/env python
"""
File       : MSPileupTasks.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: perform different set of tasks over MSPileup data
"""

# system modules
import time
import math
import asyncio

# WMCore modules
from WMCore.MicroService.Tools.Common import getMSLogger
from WMCore.MicroService.MSPileup.DataStructs.MSPileupReport import MSPileupReport
from WMCore.Services.UUIDLib import makeUUID
from WMCore.MicroService.Tools.PycurlRucio import getPileupContainerSizesRucio, getRucioToken
from WMCore.MicroService.MSPileup.MSPileupMonitoring import flatDocuments
from WMCore.MicroService.MSPileup.MSPileupError import MSPileupSchemaError
from Utils.Timers import gmtimeSeconds


class MSPileupTasks():
    """
    MSPileupTaskManager class is resposible for data placement logic. It performs
    three main tasks:
    - monitoring task to fetch current state of rule ID
    - inactive task to lookup pileup docs that has been set to inactive state
    - active task to look-up pileup docs in active state
    """

    def __init__(self, dataManager, monitManager, logger,
                 rucioAccount, rucioClient, rucioScope='cms',
                 customRucioScope='group.wmcore', dryRun=False):
        """
        MSPileupTaskManager constructor
        :param dataManager: MSPileup Data Management layer instance
        :param logger: logger instance
        :param rucioAccount: rucio account name to use
        :param rucioClient: rucio client or WMCore Rucio wrapper class to use
        :param rucioScope: rucio scope to use in rucio APIs
        :param customRucioScope: custom rucio scope to use in rucio APIs for customName DID
        :param dryRun: dry-run mode of operations
        """
        self.mgr = dataManager
        self.monitManager = monitManager
        self.logger = logger
        self.rucioAccount = rucioAccount
        self.rucioClient = rucioClient
        self.rucioScope = rucioScope
        self.customRucioScope = customRucioScope
        self.report = MSPileupReport()
        self.dryRun = dryRun
        if dryRun:
            self.logger.info("MSPileupTasks is set to DRY-RUN mode!")

    def pileupSizeTask(self):
        """
        Execute pileup size update task
        :return: a tuple (cmsDict, cusDict) which defines cms/custom containers and their
        respective dataset sizes dicts.
        """
        cmsDict = {}
        cusDict = {}
        self.logger.info("====> Executing pileupSizeTask method...")
        try:
            # get pileup sizes and update them in DB
            spec = {}
            docs = self.mgr.getPileup(spec)
            rucioAuthUrl = self.rucioClient.rucioParams.get('auth_host', '')
            rucioHostUrl = self.rucioClient.rucioParams.get('rucio_host', '')
            rucioToken, tokenValidity = getRucioToken(rucioAuthUrl, self.rucioAccount)
            msg = f"Rucio token {rucioToken} with validity {tokenValidity}"
            self.logger.info(msg)
            rucioScope = self.rucioScope
            cmsContainers = []
            customContainers = []
            datasetSizes = {}
            for doc in docs:
                if doc.get('customName', '') != '':
                    customContainers.append(doc['customName'])
                else:
                    cmsContainers.append(doc['pileupName'])
            msg = f"Fetching pileup size for {len(cmsContainers)} cms containers "
            msg += f"and {len(customContainers)} custom containers "
            msg += f"against rucio url: {rucioHostUrl}"
            self.logger.info(msg)
            if len(cmsContainers) > 0:
                datasetSizes = getPileupContainerSizesRucio(cmsContainers, rucioHostUrl, rucioToken)
                cmsDict = datasetSizes
            if len(customContainers) > 0:
                rucioScope = spec.get('customRucioScope', 'group.wmcore')
                cusDict = getPileupContainerSizesRucio(customContainers, rucioHostUrl, rucioToken, scope=rucioScope)
                datasetSizes.update(cusDict)
            for doc in docs:
                pileupName = doc.get('customName', '')
                if pileupName == '':
                    pileupName = doc['pileupName']
                pileupSize = datasetSizes.get(pileupName, 0)
                if pileupSize:
                    doc['pileupSize'] = pileupSize
                # here we update our new document in database with document validation
                self.mgr.updatePileupDocumentInDatabase(doc)
        except Exception as exp:
            msg = f"MSPileup pileup size task failed with error {exp}"
            self.logger.exception(msg)
        return cmsDict, cusDict

    def getPileupBlocks(self, doc, previousFraction, fraction):
        """
        Return list of blocks for increase/decrease scenarios of partial pileup
        according to logic outlined in
        https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5#logic-for-increasingdecreasing-container-fraction
        :param doc: pileup document
        :param previousFraction: container fraction of previous step transition
        :param fraction: current container fraction
        :return: list of blocks
        """
        if previousFraction < fraction:
            return self.getIncreasingBlocks(doc, fraction)
        return self.getDecreasingBlocks(doc, fraction)

    def getIncreasingBlocks(self, doc, fraction):
        """
        Return list of blocks for increase scenarios of partial pileup
        according to logic outlined in
        https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5#logic-for-increasingdecreasing-container-fraction
        NOTE: in order to increaee pileup fraction we already should have
        in place transition record since MSPilupeObj always created with container fraction 1

        :param doc: pileup document
        :param fraction: container fraction
        :return: list of blocks
        """
        pname = doc['pileupName']
        cname = doc['customName']
        totalBlocks = self.rucioClient.getBlocksInContainer(pname)
        customBlocks = self.rucioClient.getBlocksInContainer(cname, scope=self.customRucioScope)
        portion = math.ceil(fraction * len(totalBlocks))
        blockList = customBlocks + [b for b in totalBlocks if b not in customBlocks]
        self.logger.info("increase scenario: use %d blocks out of %d custom blocks from %s and %d from %s",
                         len(blockList), len(customBlocks), cname, abs(portion - len(customBlocks)), pname)
        return blockList[:portion]

    def getDecreasingBlocks(self, doc, fraction):
        """
        Return list of blocks for decrease scenarios of partial pileup
        according to logic outlined in
        https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5#logic-for-increasingdecreasing-container-fraction

        :param doc: pileup document
        :param fraction: container fraction
        :return: list of blocks
        """
        pname = doc['pileupName']
        cname = doc['customName']
        totalBlocks = self.rucioClient.getBlocksInContainer(pname)
        portion = math.ceil(fraction * len(totalBlocks))
        if cname == '':
            self.logger.info("decrease scenario: use %d blocks out of pileup %s with %d blocks",
                             len(totalBlocks[:portion]), pname, len(totalBlocks))
            return totalBlocks[:portion]
        customBlocks = self.rucioClient.getBlocksInContainer(cname, scope=self.customRucioScope)
        self.logger.info("decrease scenario: use %d blocks out of custom container %s",
                         portion, cname)
        return customBlocks[:portion]

    def partialPileupTask(self):
        """
        Execute partial pileup placement according to the following logic:

        - get rucio DIDs (datasets, in CMS name blocks) for our pileup document
        - get portion of DIDs based on ceil(containerFraction * num_rucio_datasets)
        - create new container DID in Rucio via Rucio wrapper createContainer API
          which calls rucio CLI add_container
        - update MSPileup JSON with
          - create new custom Name as pileup+extention
          - add new transition record and update MSPileup document
        - we call attachDIDs Rucio wrapper API with our set of DIDs and rses from pileup document
        - create new rules for custom DID via provided rseExpression
        - add new ruleIds to pileup document
        - set expiration date (to be 24h) for already existing ruleIds from pileup document

        :return: None
        """
        spec = {'active': True}
        docs = self.mgr.getPileup(spec)
        for doc in docs:
            self.logger.info("partialPileupTask process pileup name %s", doc['pileupName'])
            fraction = doc['containerFraction']

            # check previous fraction value and proceed only if has changed
            transition = doc.get('transition', [])
            if transition:
                previousFraction = transition[-1]['containerFraction']
            else:
                previousFraction = doc['containerFraction']
            if previousFraction == fraction:
                # do nothing since there is no transition, i.e. skip this record
                self.logger.info("Pileup name %s has no container fractions changes", doc['pileupName'])
                continue

            # get rucio DIDs (datasets, in CMS name blocks) for our pileup document as following:
            # - when container fraction is decreasing: new container is a subset of customName
            # - when container fraction is increasing: new container is a superset of customName + subset of pileupName

            # usage of block names defined in this logic:
            # https://github.com/dmwm/WMCore/pull/11807#pullrequestreview-1786778783
            customBlocks = self.getPileupBlocks(doc, previousFraction, fraction)

            # create new container DID in Rucio for transition custom name
            cname = doc['transition'][-1]['customDID']
            self.logger.info("Create container %s with scope %s", cname, self.customRucioScope)
            status = self.rucioClient.createContainer(cname, scope=self.customRucioScope)
            if not status:
                self.logger.error("Failed to create container %s with scope %s", cname, self.customRucioScope)
                continue

            # call rucio APIs to attach custom blocks to our custom container (DID)
            newRules = []
            self.logger.info("Attaching %d blocks to custom pileup name: %s", len(customBlocks), cname)
            if len(customBlocks):
                self.rucioClient.attachDIDs(None, cname, customBlocks, scope=self.customRucioScope)
            for rse in doc['expectedRSEs']:
                # create new rule for custom DID using pileup document rse
                ruleIds = self.rucioClient.createReplicationRule(cname, rse, scope=self.customRucioScope)
                self.logger.info("Rule ids: %s created for custom pileup: %s for RSE: %s", ruleIds, cname, rse)
                newRules += ruleIds

            self.logger.info("Custom pileup: %s has the following new rules created: %s for RSEs: %s",
                             cname, newRules, doc['expectedRSEs'])
            # set expiration date (to be 24h) for already existing ruleIds from pileup document
            for rid in doc['ruleIds']:
                # set expiration date to be 24h ahead of right now
                opts = {'lifetime': 24 * 60 * 60}
                self.rucioClient.updateRule(rid, opts)
                self.logger.info("Rule id: %s has been updated with lifetime: %s", rid, opts['lifetime'])

            # update pileup document
            if newRules:
                # remove duplicate rules
                newRules = list(set(newRules))
                try:
                    # update rules
                    doc['ruleIds'] = newRules

                    # update transition record of the pileup document, see logic:
                    # https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5

                    # get previous transition record
                    prevTranRecord = doc['transition'][-1]

                    # keep custom name value from this record as we'll update it later
                    cname = prevTranRecord['customDID']

                    # update previous transition record in place, i.e. it will be updated
                    # within doc['transition'] directly after these assingments
                    prevTranRecord['containerFraction'] = fraction
                    prevTranRecord['updateTime'] = gmtimeSeconds()

                    # update custom name from value of previous transition record
                    doc['customName'] = cname
                    self.logger.info("Finally, updating the pileup document for pileup name: %s", doc['pileupName'])

                    # update MSPileup document in MongoDB
                    # here we update our new document in database with document validation
                    self.mgr.updatePileupDocumentInDatabase(doc, rseList=doc['currentRSEs'])
                    self.logger.info("Pileup name %s had its fraction updated in the partialPileupTask function", doc['pileupName'])
                except Exception as exp:
                    msg = f"Failed to update MSPileup document, {exp}"
                    self.logger.exception(msg)
                    err = MSPileupSchemaError(doc, msg)
                    self.logger.error(err)

    def cleanupTask(self, cleanupDaysThreshold):
        """
        Execute cleanup task according to the following logic:
        1. Fetch documents from backend database for the following conditions
           - active=False, and
           - empty ruleIds list; and
           - empty currentRSEs; and
           document has been deactivated for a while (deactivatedOn=XXX),
        2. For those documents which are fetched make delete call to backend database
        :param timeThreshold: time threshold in days which will determine document clean-up readiness
        """
        self.logger.info("====> Executing cleanupTask method...")
        spec = {'active': False}
        docs = self.mgr.getPileup(spec)
        deleteDocs = 0
        seconds = cleanupDaysThreshold * 24 * 60 * 60  # convert to second
        for doc in docs:
            pileupName = doc.get('customName', '')
            if pileupName == '':
                pileupName = doc['pileupName']
            if not doc['ruleIds'] and not doc['currentRSEs'] and \
                    time.time() > doc['deactivatedOn'] + seconds:
                spec = {'pileupName': pileupName}
                if not self.dryRun:
                    self.logger.info("Cleanup task deleting pileup %s", pileupName)
                    self.mgr.deletePileup(spec)
                    deleteDocs += 1
                else:
                    self.logger.info(f"DRY-RUN: should have deleted pileup document for {pileupName}")
        self.logger.info("Cleanup task deleted %d pileup objects", deleteDocs)

    def cmsMonitTask(self):
        """
        Execute CMS MONIT task according to the following logic:

        1. Read all pileup document from MongoDB
        2. Flatten all docs
        3. Submit flatten docs to CMS MONIT
        """
        self.logger.info("====> Executing cmsMonitTask method...")
        if not self.monitManager.userAMQ or not self.monitManager.passAMQ:
            self.logger.info("MSPileupMonitoring has no AMQ credentials, will skip the upload to MONIT")
            return
        spec = {}
        msPileupDocs = self.mgr.getPileup(spec)
        docs = []
        for doc in msPileupDocs:
            for flatDoc in flatDocuments(doc):
                docs.append(flatDoc)
        results = self.monitManager.uploadToAMQ(docs)
        if results and isinstance(results, dict):
            success = results['success']
            failures = results['failures']
            msg = f"MSPileup CMS MONIT task fetched {len(msPileupDocs)} docs from MSPileup backend DB"
            msg += f", and sent {len(docs)} flatten docs to MONIT"
            msg += f", number of success docs {success} and failures {failures},"
            self.logger.info(msg)
        else:
            self.logger.error("MSPileup CMS MONIT task failed!")

    def monitoringTask(self):
        """
        Execute Monitoring task according to the following logic:

        1. Read pileup document from MongoDB with filter active=true
        2. For each rule id in ruleIds:
           - query Rucio for that rule id and fetch its state (e.g.: afd122143kjmdskj)
           - if state=OK, log that the rule has been satisfied and add that RSE to
           the currentRSEs (unique)
           - otherwise, calculate the rule completion based on the 3 locks_* field
        3. now that all the known rules have been inspected, persist the up-to-date
        pileup doc in MongoDB
        """
        self.logger.info("====> Executing monitoringTask method...")
        spec = {'active': True}
        docs = self.mgr.getPileup(spec)
        taskSpec = self.getTaskSpec()
        self.logger.info("Running the monitoring task on %d pileup objects", len(docs))
        asyncio.run(performTasks('monitoring', docs, taskSpec))

    def inactiveTask(self):
        """
        Inactive pileup task:

        This task is supposed to look at pileup documents that have been set to
        inactive. The main goal here is to ensure that there are no Rucio rules
        left in the system (of course, for the relevant DID and the Rucio
        account adopted by our microservice). Pileup documents that are updated
        as a result of this logic should have their data persisted back in
        MongoDB. A short algorithm for it can be done described as follows:

        1. Read pileup document from MongoDB with filter active=false
        2. for each DID and Rucio account, get a list of all the existent rules
           - make a Rucio call to delete that rule id, then:
             - remove the rule id from ruleIds (if any) and remove the RSE name
             from currentRSEs (if any)
        3. make a log record if the DID + Rucio account tuple does not have any
        existent rules
           - and set ruleIds and currentRSEs to an empty list
        4. once all the relevant rules have been removed, persist an up-to-date
        version of the pileup data structure in MongoDB
        """
        self.logger.info("====> Executing inactiveTask method...")
        spec = {'active': False}
        docs = self.mgr.getPileup(spec)
        taskSpec = self.getTaskSpec()
        self.logger.info("Running the inactive task on %d pileup objects", len(docs))
        asyncio.run(performTasks('inactive', docs, taskSpec))

    def activeTask(self, marginSpace=1024**4):
        """
        Active pileup task:
        :param marginSpace: minimum margin space size in bytes to have at RSE to place a dataset

        This task is supposed to look at pileup documents active in the system.
        Its main goal is to ensure that the pileup DID has all the requested
        rules (and nothing beyond them), according to the pileup object
        configuration. Pileup documents that are updated as a result of this
        logic should have their data persisted back in MongoDB.

        1. Read pileup document from MongoDB with filter active=true
        2. if expectedRSEs is different than currentRSEs, then further data placement
        is required (it's possible that data removal is required!)
        3. make a local copy of the currentRSEs value to track rules incomplete but ongoing
        4. for each rule matching the DID + Rucio account, perform:
           - if rule RSE is not in expectedRSEs, then this rule needs to be deleted.
             - upon successful rule deletion, also remove the RSE name
             from currentRSEs and ruleIds (if any)
             - make a log record
           - else, save this rule RSE in the local copy of currentRSEs.
           This rule is likely still being processed by Rucio.
        5. now that we evaluated expected versus current,
        for each expectedRSEs not in our local copy of currentRSEs:
           - first, check whether the RSE has enough available space available for that
           (we can assume that any storage with less than 1 TB available cannot be
           considered for pileup data placement)
             - in case of no space available, make a log record
           - in case there is enough space, make a Rucio rule for
           that DID + Rucio account + RSE
             - now append the rule id to the ruleIds list
        6. once all the relevant rules have been created,
        or if there was any changes to the pileup object,
        persist an up-to-date version of the pileup data structure in MongoDB
        """
        self.logger.info("====> Executing activeTask method...")
        spec = {'active': True}
        docs = self.mgr.getPileup(spec)
        taskSpec = self.getTaskSpec()
        taskSpec['marginSpace'] = marginSpace
        self.logger.info("Running the active task on %d pileup objects", len(docs))
        asyncio.run(performTasks('active', docs, taskSpec))

    def getTaskSpec(self):
        """Return task spec"""
        spec = {'manager': self.mgr, 'logger': self.logger, 'report': self.report,
                'rucioClient': self.rucioClient, 'rucioAccount': self.rucioAccount,
                'rucioScope': self.rucioScope, 'customRucioScope': self.customRucioScope,
                'dryRun': self.dryRun}
        return spec

    def getReport(self):
        """
        Return report object to upstream codebase
        """
        return self.report


def monitoringTask(doc, spec):
    """
    Perform single monitoring task over provided MSPileup document

    :param doc: MSPileup document
    :param spec: task spec dict
    """
    mgr = spec['manager']
    uuid = spec['uuid']
    rucioClient = spec['rucioClient']
    rucioAccount = spec['rucioAccount']
    rucioScope = spec['rucioScope']
    report = spec['report']
    logger = spec['logger']

    # get list of existent rule ids
    pname = doc.get('customName', '')
    if pname == '':
        pname = doc['pileupName']
    else:
        rucioScope = spec.get('customRucioScope', 'group.wmcore')
    kwargs = {'scope': rucioScope, 'account': rucioAccount}
    rules = rucioClient.listDataRules(pname, **kwargs)
    modify = False

    if not rules:
        logger.info(f"Did not find any {rucioAccount} rules for container: {pname}.")
        if not doc['expectedRSEs']:
            logger.warning(f"Container: {pname} is active but has no expected RSEs.")
        elif doc['currentRSEs'] or doc['ruleIds']:
            doc['currentRSEs'].clear()
            doc['ruleIds'].clear()
            modify = True

    for rdoc in rules:
        rses = rucioClient.evaluateRSEExpression(rdoc['rse_expression'])
        rid = rdoc['id']
        state = rdoc['state']

        # rucio state have the following values
        # https://github.com/rucio/rucio/blob/master/lib/rucio/db/sqla/constants.py
        # the states are: OK, REPLICATING, STUCK, SUSPENDED, WAITING_APPROVAL, INJECT
        if state == 'OK':
            # log that the rule has been satisfied and add that RSE to the currentRSEs (unique)
            msg = f"monitoring task {uuid}, container {pname} under the rule ID {rid}"
            msg += f" targeting RSEs {rses} in rucio account {rucioAccount} is completely available"
            logger.info(msg)
            for rse in rses:
                if rse not in doc['currentRSEs']:
                    doc['currentRSEs'].append(rse)
                    modify = True
                    msg = f"update currentRSEs with {rse}"
                    report.addEntry('monitoring', uuid, msg)
        else:
            # calculate the rule completion based on the 3 locks_* field
            sumOfLocks = rdoc['locks_ok_cnt'] + rdoc['locks_replicating_cnt'] + rdoc['locks_stuck_cnt']
            completion = rdoc['locks_ok_cnt'] / sumOfLocks
            msg = f"monitoring task {uuid}, container {pname} under the rule ID {rid}"
            msg += f" targeting RSEs {rses} in rucio account {rucioAccount} has"
            msg += f" a fraction completion of {completion}"
            logger.info(msg)
            for rse in rses:
                if rse in doc['currentRSEs']:
                    doc['currentRSEs'].remove(rse)
                    modify = True
                    msg = f"delete rse {rse} from currentRSEs"
                    report.addEntry('monitoring', uuid, msg)

        # now keep track of this rule id in the pileup document
        for rse in rses:
            if rse in doc['expectedRSEs'] and rid not in doc['ruleIds']:
                logger.info(f"Tracking rule id {rid} that was not created by MSPileup")
                doc['ruleIds'].append(rid)
                modify = True
                break

    # persist an up-to-date version of the pileup data structure in MongoDB
    if modify:
        logger.info(f"monitoring task {uuid}, update {pname}")
        # here we update our new document in database without validation since
        # we only updated rule ids and rses
        mgr.updatePileupDocumentInDatabase(doc, validate=False)
        msg = f"update pileup {pname}"
        report.addEntry('monitoring', uuid, msg)
    else:
        logger.info(f"monitoring task {uuid}, processed without update for {pname}")


def inactiveTask(doc, spec):
    """
    Perform single inactive task over provided MSPileup document

    :param doc: MSPileup document
    :param spec: task spec dict
    """
    mgr = spec['manager']
    uuid = spec['uuid']
    rucioClient = spec['rucioClient']
    rucioAccount = spec['rucioAccount']
    rucioScope = spec['rucioScope']
    report = spec['report']
    logger = spec['logger']
    dryRun = spec['dryRun']
    pname = doc.get('customName', '')
    if pname == '':
        pname = doc['pileupName']
    else:
        rucioScope = spec.get('customRucioScope', 'group.wmcore')
    kwargs = {'scope': rucioScope, 'account': rucioAccount}
    rules = rucioClient.listDataRules(pname, **kwargs)
    modify = False

    for rdoc in rules:
        # make a Rucio call to delete that rule id
        rid = rdoc['id']
        msg = f"inactive task {uuid}, container: {pname} for Rucio account {rucioAccount}"
        msg += f", delete replication rule {rid}"
        logger.info(msg)
        if not dryRun:
            rucioClient.deleteRule(rid)
        else:
            logger.info(f"DRY-RUN: rule id '{rid}' should have been deleted.")
            continue
        rses = rucioClient.evaluateRSEExpression(rdoc['rse_expression'])

        # remove the rule id from ruleIds (if any) and remove the RSE name from currentRSEs (if any)
        if rid in doc['ruleIds']:
            doc['ruleIds'].remove(rid)
            modify = True
            msg = f"remove rid {rid}"
            report.addEntry('inactive', uuid, msg)
        for rse in rses:
            if rse in doc['currentRSEs']:
                doc['currentRSEs'].remove(rse)
                modify = True
                msg = f"remove rse {rse}"
                report.addEntry('inactive', uuid, msg)

    # make a log record if the DID + Rucio account tuple does not have any existent rules
    if not rules:
        msg = f"inactive task {uuid}, container: {pname} for Rucio account {rucioAccount}"
        msg += " does not have any existing rules, proceed without update"
        logger.info(msg)

    # persist an up-to-date version of the pileup data structure in MongoDB
    if modify:
        logger.info(f"inactive task {uuid}, update {pname}")
        # here we update our new document in database without validation since
        # we only updated rule ids and rses
        mgr.updatePileupDocumentInDatabase(doc, validate=False)
        msg = f"update pileup {pname}"
        report.addEntry('inactive', uuid, msg)


def activeTask(doc, spec):
    """
    Perform single active task over provided MSPileup document

    :param doc: MSPileup document
    :param spec: task spec dict
    """
    mgr = spec['manager']
    uuid = spec['uuid']
    logger = spec['logger']
    rucioClient = spec['rucioClient']
    rucioAccount = spec['rucioAccount']
    rucioScope = spec['rucioScope']
    report = spec['report']
    marginSpace = spec['marginSpace']
    dryRun = spec['dryRun']

    # extract relevant part of our pileup document we'll use in our logic
    expectedRSEs = set(doc['expectedRSEs'])
    currentRSEs = set(doc['currentRSEs'])
    pileupSize = doc['pileupSize']
    pname = doc.get('customName', '')
    if pname == '':
        pname = doc['pileupName']
    else:
        rucioScope = spec.get('customRucioScope', 'group.wmcore')
    inputArgs = {'scope': rucioScope, 'name': pname, 'account': rucioAccount}
    kwargs = {'scope': rucioScope, 'account': rucioAccount}
    modify = False
    localCopyOfCurrentRSEs = list(currentRSEs)

    if not pileupSize:
        logger.error(f"Pileup {pname} does not have its size defined: {pileupSize}")
    elif expectedRSEs != currentRSEs:
        # this means that further data placement is required, or it could be
        # that that data removal is actually required!
        msg = f"active task {uuid}"
        msg += f", further data placement required for pileup name: {pname},"
        msg += f" with expectedRSEs: {expectedRSEs} and data currently available at: {currentRSEs}"
        logger.info(msg)

        # get list of replication rules for our scope, pileup name and account
        rules = rucioClient.listDataRules(pname, **kwargs)
        # for each rse_expression in rules get list of rses
        for rdoc in rules:
            rses = rucioClient.evaluateRSEExpression(rdoc['rse_expression'])
            rdoc['rses'] = rses

        # for each rule matching the DID + Rucio account
        for rdoc in rules:
            rid = rdoc['id']
            rses = rdoc['rses']
            for rse in rses:
                # if rule RSE is not in expectedRSEs, then this rule needs to be deleted
                if rse not in expectedRSEs:
                    # upon successful rule deletion, also remove the RSE name from currentRSEs
                    if not dryRun:
                        rucioClient.deleteRule(rid)
                    else:
                        logger.info(f"DRY-RUN: rule id '{rid}' should have been deleted.")
                    if rse in doc['currentRSEs']:
                        doc['currentRSEs'].remove(rse)
                        modify = True
                    msg = f"rse {rse} rule is deleted and remove from currentRSEs of {pname}"
                    logger.info(msg)
                    # delete rid in ruleIds (if any)
                    if rid in doc['ruleIds']:
                        doc['ruleIds'].remove(rid)
                        modify = True
                        msg = f"rule Id {rid} is removed from ruleIds for {pname}"
                        logger.info(msg)
                else:
                    # else, save this rule RSE in the local copy of currentRSEs.
                    # This rule is likely still being processed by Rucio.
                    if rse not in localCopyOfCurrentRSEs:
                        localCopyOfCurrentRSEs.append(rse)
                        msg = f"rse {rse} is added to localCopyOfCurrentRSEs"
                        logger.info(msg)

        # for each expectedRSEs not in localCopyOfCurrentRSEs
        for rse in set(expectedRSEs).difference(localCopyOfCurrentRSEs):
            # check whether the RSE has enough available space available for that
            # taken from logic of RSEQuotas, see https://bit.ly/3kDJZmO
            # but here we use explicitly rse parameter
            records = rucioClient.getAccountUsage(rucioAccount, rse)
            enoughSpace = False
            for rec in records:
                # each record has form
                # {"rse": ..., "bytes_limit": ..., "bytes": ..., "bytes_remaining": ...}
                if rec['bytes_remaining'] - pileupSize > marginSpace:
                    enoughSpace = True
                    break
            dids = [inputArgs]
            if enoughSpace:
                msg = f"active task {uuid}, for dids {dids} there is enough space at RSE {rse}"
                logger.warning(msg)
                # create the rule and append the rule id to ruleIds
                if not dryRun:
                    rids = rucioClient.createReplicationRule(pname, rse, **kwargs)
                else:
                    msg = f"DRY-RUN: rule for pileup {pname} and rse {rse} should have been created"
                    logger.info(msg)
                    continue
                # add new ruleId to document ruleIds
                for rid in rids:
                    if rid not in doc['ruleIds']:
                        doc['ruleIds'].append(rid)
                        modify = True
            else:
                # make a log record saying that there is not enough space
                msg = f"active task {uuid}, for dids {dids} there is not enough space at RSE {rse}"
                logger.warning(msg)

    # persist an up-to-date version of the pileup data structure in MongoDB
    if modify:
        logger.info(f"active task {uuid} update {pname}")
        # here we update our new document in database without validation since
        # we only updated rule ids and rses
        mgr.updatePileupDocumentInDatabase(doc, validate=False)
        msg = f"update pileup {pname}"
        report.addEntry('active', uuid, msg)
    else:
        logger.info(f"active task {uuid}, processed without update")


async def runTask(task, doc, spec):
    """
    Run specified task for given document and spec

    :param task: task to perform
    :param doc: MSPileup document to process
    :param spec: task spec dictionary
    """
    time0 = time.time()
    report = spec['report']
    logger = spec.get('logger', getMSLogger(False))
    pname = doc.get('customName', '')
    if pname == '':
        pname = doc['pileupName']

    # set report hash
    uuid = makeUUID()
    spec['uuid'] = uuid
    report.addEntry(task, uuid, 'starts')
    msg = f"MSPileup {task} task {uuid} pileup {pname}"
    try:
        if task == 'monitoring':
            monitoringTask(doc, spec)
        elif task == 'inactive':
            inactiveTask(doc, spec)
        elif task == 'active':
            activeTask(doc, spec)
        msg += ", successfully processed"
    except Exception as exp:
        msg += f", failed with error {exp}"
        logger.exception(msg)
    report.addEntry(task, uuid, msg)

    # update task report
    etime = time.time() - time0
    msg = "ends, elapsed time %.2f (sec)" % etime
    report.addEntry(task, uuid, msg)


async def performTasks(task, docs, spec):
    """
    Perform tasks via async IO co-routines

    :param task: task to perform
    :param docs: list of MSPileup documents to process
    :param spec: task spec dictionary
    """
    coRoutines = [runTask(task, doc, spec) for doc in docs]
    await asyncio.gather(*coRoutines)



#!/usr/bin/env python
"""
    WorkQueueElement unit tests
"""
from __future__ import (print_function, division)

import json
import os
import unittest

from WMCore.WMBase import getTestBase
from WMCore.WorkQueue.DataStructs.WorkQueueElement import WorkQueueElement
from WMCore.WorkQueue.DataStructs.WorkQueueElementsSummary import WorkQueueElementsSummary


class WorkQueueElementsSummaryTest(unittest.TestCase):

    def setUp(self):


        filePath = os.path.join(getTestBase(),
                                    "WMCore_t/WorkQueue_t/DataStructs_t/wq_available_elements.json")
        with open(filePath, "r") as f:
            gqData = json.load(f)

        self.gqElements = []
        for ele in gqData:
            self.gqElements.append(WorkQueueElement(**ele))



    def testElementsWithHigherPriorityInSameSites(self):
        """Id calculated correctly"""

        gqSummary = WorkQueueElementsSummary(self.gqElements)
        testReq = "riahi_TEST_HELIX_0911-T1_UK_RALBackfill_151119_190209_6101"
        filteredElements = gqSummary.elementsWithHigherPriorityInSameSites(testReq, returnFormat="list")

        wqSummary = WorkQueueElementsSummary(filteredElements)
        wqElements = wqSummary.getWQElementResultsByRequest()

        self.assertEqual(gqSummary.getPossibleSitesByRequest(testReq), set(['T1_UK_RAL']))
        priority = gqSummary.getWQElementResultsByRequest(testReq)['Priority']
        self.assertEqual(priority, 50000)

        self.assertEqual(len(wqElements), 2)

        jobs = 0
        for req in wqElements:
            jobs += wqElements[req]['Jobs']
            self.assertTrue(wqElements[req]['Priority'] >= priority)
        self.assertEqual(jobs, 1942)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_Sandbox_

Sandbox access API to lookup and retrieve files from the sandbox area
at runtime.

Sandbox is found by looking for WMSandbox module and working down
the module tree imposed by it


"""

from builtins import object
import os
import inspect

class Sandbox(object):
    """
    _Sandbox_


    """
    def __init__(self, taskName, stepName):
        self.moduleName = "WMSandbox.%s.%s" % (taskName, stepName)
        self.module = __import__(self.moduleName,
                                 globals(), locals(), [stepName])
                                 #globals(), locals(), [stepName], -1)
        self.directory = os.path.dirname(inspect.getsourcefile(self.module))


    def listFiles(self):
        """
        _listFiles_

        List files in sandbox for this instance

        """
        allFiles = [ x for x in os.listdir(self.directory)
                     if not x.startswith("__init__") ]
        return allFiles



    def getFile(self, filename):
        """
        _getFile_

        Get the absolute location of the file in the sandbox

        """
        if filename not in self.listFiles():
            msg = "Cannot find file %s in %s" % (filename, self.directory)
            #raise or return None here??
            raise RuntimeError(msg)
        return os.path.join(self.directory, filename)



import unittest

import WMCore.Storage.Backends
import WMCore.Storage.Plugins.TestWinImpl
from WMCore.Storage.Registry import retrieveStageOutImpl


class TestRegistry(unittest.TestCase):
    def testRetrieveProperBackends(self):
        # make sure the default still gives the old behavior
        self.assertTrue(isinstance(retrieveStageOutImpl('test-win'),
                                   WMCore.Storage.Backends.UnittestImpl.WinImpl))
        # make sure you can still explicitly get the old one
        self.assertTrue(isinstance(retrieveStageOutImpl('test-win', useNewVersion=False),
                                   WMCore.Storage.Backends.UnittestImpl.WinImpl))
        # make sure that you get the new one if you want it
        self.assertTrue(retrieveStageOutImpl('test-win', useNewVersion=True),
                        WMCore.Storage.Plugins.TestWinImpl.TestWinImpl)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

from __future__ import (division, print_function)

endpoint = 'https://cmsweb-prod.cern.ch/dbs/prod/phys03/DBSReader'

datasets = []

calls = [
    ['listBlockOrigin', {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caabcd'}],
    ['listBlockOrigin', {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace'}],
    ['listBlockOrigin', {'block_name': '/HighPileUp/Run2011A-v1/RAW#6021175e-cbfb-11e0-80a9-003048caaace'}],
    ['listBlockOrigin', {'block_name': '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caabcdabcd'}],
    ['listBlockOrigin', {
        'block_name': '/GenericTTbar/hernan-140317_231446_crab_JH_ASO_test_T2_ES_CIEMAT_5000_100_140318_0014-ea0972193530f531086947d06eb0f121/USER#fb978442-a61b-413a-b4f4-526e6cdb142e'}],
    ['listBlockOrigin', {
        'block_name': '/GenericTTbar/hernan-140317_231446_crab_JH_ASO_test_T2_ES_CIEMAT_5000_100_140318_0014-ea0972193530f531086947d06eb0f121/USER#0b04d417-d734-4ef2-88b0-392c48254dab'}],
]



#!/usr/bin/env python
"""
_Proxy_t_
Test the basic Proxy operations.
You need to source your UI before running these tests.
Your proxy is initialized in testAAACreateProxy method and it is used by the remaining tests.
"""
from __future__ import print_function
from __future__ import division

from past.utils import old_div
import unittest
import os
import logging
import logging.config
import time
import subprocess

from nose.plugins.attrib import attr
from WMCore.Credential.Proxy import Proxy
from Utils.PythonVersion import PY3
from Utils.Utilities import decodeBytesToUnicode

# You may have to set these environment variables to run in a local environment

group = os.environ.get('PROXY_GROUP', '')
role = os.environ.get('PROXY_ROLE', 'NULL')
myProxySvr = os.environ.get('MYPROXY_SERVER', 'myproxy.cern.ch')

class ProxyTest(unittest.TestCase):

    def setUp(self):
        """
        Setup for unit tests
        """
        logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M',
                    filename='proxy_unittests.log',
                    filemode='w')

        logger_name = 'ProxyTest'

        self.logger = logging.getLogger(logger_name)
        self.dict = {'logger': self.logger,
                     'vo': 'cms', 'group': group, 'role': role, 'myProxySvr': myProxySvr,
                     'proxyValidity' : '192:00', 'min_time_left' : 36000}

        self.proxyPath = None
        self.proxy = Proxy( self.dict )

    def tearDown(self):
        """
        _tearDown_

        Tear down the proxy.
        """
        return

    def getUserIdentity(self):
        """
        _getUserIdentity_
        Retrieve the user's subject from the voms-proxy-info call.
        """
        vomsProxyInfoCall = subprocess.Popen(["voms-proxy-info", "-identity"],
                                             stdout = subprocess.PIPE,
                                             stderr = subprocess.PIPE)
        if vomsProxyInfoCall.wait() != 0:
            return None

        stdout, _ = vomsProxyInfoCall.communicate()
        stdout = decodeBytesToUnicode(stdout) if PY3 else stdout
        return stdout[0:-1]

    def getUserAttributes(self):
        """
        _getUserAttributes_
        Retrieve the user's attributes from the voms-proxy-info call.
        """
        vomsProxyInfoCall = subprocess.Popen(["voms-proxy-info", "-fqan"],
                                             stdout = subprocess.PIPE,
                                             stderr = subprocess.PIPE)
        if vomsProxyInfoCall.wait() != 0:
            return None

        stdout, _ = vomsProxyInfoCall.communicate()
        stdout = decodeBytesToUnicode(stdout) if PY3 else stdout
        return stdout[0:-1]

    @attr("integration")
    def testGetUserCertEnddate( self ):
        """
        Test if getTimeLeft method returns correctly the proxy time left.
        """
        daysleft = self.proxy.getUserCertEnddate()
        self.assertEqual(daysleft, 58) #set this as the number of days left in .globus/usercert.pem
        daysleft = self.proxy.getUserCertEnddate(openSSL=False)
        self.assertEqual(daysleft, 58) #set this as the number of days left in .globus/usercert.pem

    @attr("integration")
    def testAAACreateProxy( self ):
        """
        Test if create method creates correctly the proxy.
        This is sort of bad form to require that this test run first, but the alternative is
        entering a password for every single invocation
        """
        self.proxy.create()
        time.sleep( 5 )
        proxyPath = self.proxy.getProxyFilename()
        self.assertTrue(os.path.exists(proxyPath))

    @attr("integration")
    def testCheckProxyTimeLeft( self ):
        """
        Test if getTimeLeft method returns correctly the proxy time left.
        """
        timeLeft = self.proxy.getTimeLeft()
        self.assertEqual(int(int(timeLeft) // 3600), 191)

    @attr("integration")
    def testRenewProxy( self ):
        """
        Test if the renew method renews correctly the user proxy.
        """
        time.sleep( 70 )
        self.proxy.renew()
        time.sleep( 10 )
        timeLeft = self.proxy.getTimeLeft()
        self.assertEqual(int(int(timeLeft) // 3600), 191)

    @attr("integration")
    def testDestroyProxy(self ):
        """
        Test the proxy destroy method.
        """
        self.proxy.destroy( )
        self.proxyPath = self.proxy.getProxyFilename()
        self.assertFalse(os.path.exists(self.proxyPath))
        # Create the proxy after the destroy
        self.proxy.create()

    @attr("integration")
    def testGetSubject(self):
        """
        _testGetSubject_
        Verify that the getSubject() method works correctly.
        """
        subject = self.proxy.getSubject( )
        self.assertEqual(subject, self.getUserIdentity(),
                         "Error: Wrong subject.")
        return

    @attr("integration")
    def testGetUserName( self ):
        """
        _testGetUserName_
        Verify that the getUserName() method correctly determines the user's
        name.
        """
        user = self.proxy.getUserName( )
        identity = self.getUserIdentity().split("/")[ len(self.getUserIdentity().split("/")) - 1 ][3:]
        self.assertEqual(user, identity,
                         "Error: User name is wrong: |%s|\n|%s|" % (user, identity))
        return

    @attr("integration")
    def testCheckAttribute( self ):
        """
        Test if the checkAttribute  method checks correctly the attributes validity.
        """
        valid = self.proxy.checkAttribute( )
        self.assertTrue(valid)

    @attr("integration")
    def testCheckTimeLeft( self ):
        """
        Test if the check method checks correctly the proxy validity.
        """
        valid = self.proxy.check( self.proxyPath )
        self.assertTrue(valid)

    @attr("integration")
    def testVomsRenewal( self ):
        """
        Test if vomsExtensionRenewal method renews correctly the voms-proxy.
        """
        proxyPath = self.proxy.getProxyFilename( )
        time.sleep( 70 )
        attribute = self.proxy.prepareAttForVomsRenewal( self.proxy.getAttributeFromProxy( proxyPath ) )
        self.proxy.vomsExtensionRenewal( proxyPath, attribute )
        vomsTimeLeft = self.proxy.getVomsLife( proxyPath )
        self.assertEqual(int(int(vomsTimeLeft) // 3600), 191)

    @attr("integration")
    def testElevateAttribute( self ):
        """
        Test if the vomsExtensionRenewal method elevate last attributes given.
        """
        proxyPath = self.proxy.getProxyFilename( )
        attribute = self.proxy.prepareAttForVomsRenewal( '/cms/Role=NULL/Capability=NULL' )
        self.proxy.vomsExtensionRenewal( proxyPath, attribute )
        self.assertEqual(self.proxy.getAttributeFromProxy(proxyPath), '/cms/Role=NULL/Capability=NULL')
        # Restore the original configuration of the proxy
        self.proxy.create()

    @attr("integration")
    def testUserGroupInProxy( self ):
        """
        Test if getUserAttributes method returns correctly the user group.
        """
        self.assertTrue(self.proxy.group, 'No group set. Testing incomplete.')
        self.assertEqual(self.proxy.group, self.getUserAttributes().split('\n')[0].split('/')[2])

    @attr("integration")
    def testUserRoleInProxy( self ):
        """
        Test if getUserAttributes method returns correctly the user role.
        """
        self.assertEqual(self.proxy.role, self.getUserAttributes().split('\n')[0].split('/')[3].split('=')[1])

    @attr("integration")
    def testGetAttributes( self ):
        """
        Test getAttributeFromProxy method.

        Can tested this with:
            voms-proxy-init -voms cms:/cms/integration #or any group of yours
            export PROXY_GROUP=integration
            python test/python/WMCore_t/Credential_t/Proxy_t.py ProxyTest.testGetAttributes
        """
        self.assertTrue(self.proxy.group, 'No group set. Testing incomplete.')
        if not self.dict['role']:
            role = 'NULL'
        else:
            role = self.dict['role']
        proxyPath = self.proxy.getProxyFilename( )
        self.assertEqual(self.proxy.getAttributeFromProxy(proxyPath).split('/')[2], self.dict['group'])
        self.assertEqual(self.proxy.getAttributeFromProxy(proxyPath).split('/')[3].split('=')[1], role)

        #test with the allAttributes flag
        self.assertTrue(self.proxy.getAttributeFromProxy(proxyPath, allAttributes=True)>1)

    @attr("integration")
    def testGetUserGroupAndRole( self ):
        """
        Test GetUserGroupAndRoleFromProxy method.
        """
        if not self.dict['role']:
            role = 'NULL'
        else:
            role = self.dict['role']
        proxyPath = self.proxy.getProxyFilename( )
        if self.dict['group'] and self.dict['role']:
            self.assertEqual(self.proxy.getUserGroupAndRoleFromProxy(proxyPath)[0], self.dict['group'])
            self.assertEqual(self.proxy.getUserGroupAndRoleFromProxy(proxyPath)[1], role)

    @attr("integration")
    def testGetAllUserGroups( self ):
        """
        Test GetAllUserGroups method.
        """
        proxyPath = self.proxy.getProxyFilename( )
        groups = self.proxy.getAllUserGroups(proxyPath)
        print(list(groups))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Unittests for XMLUtils functions
"""

from __future__ import division, print_function

import unittest

from Utils.PythonVersion import PY3

from WMCore.Services.TagCollector.XMLUtils import xml_parser


class XMLUtilsTest(unittest.TestCase):
    """
    unittest for XMLUtils functions
    """

    def setUp(self):
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def testParser(self):
        """
        Test parsing a basic XML as the one retrieved from TagCollector
        """
        data = '<projects>\n' \
               '  <architecture name="slc5_amd64_gcc462">\n' \
               '    <project label="CMSSW_5_3_6" type="Production" state="Announced"/>\n' \
               '    <project label="CMSSW_5_3_7" type="Production" state="Announced"/>\n' \
               '    <project label="CMSSW_5_3_8" type="Production" state="Announced"/>\n' \
               '  </architecture>\n' \
               '  <architecture name="slc7_ppc64le_gcc530">\n' \
               '    <project label="CMSSW_8_1_0_pre15" type="Development" state="Announced"/>\n' \
               '  </architecture>\n' \
               '  <architecture name="slc6_amd64_gcc620">\n' \
               '    <project label="CMSSW_9_0_0_pre3" type="Development" state="Announced"/>\n' \
               '    <project label="CMSSW_9_0_0_pre4" type="Development" state="Announced"/>\n' \
               '  </architecture>\n' \
               '</projects>\n'
        pkey = 'architecture'

        for row in xml_parser(data, pkey):
            self.assertEqual(type(row[pkey]['project']), list)
            print(row)
            if row[pkey]['name'] == 'slc5_amd64_gcc462':
                releases = [item['label'] for item in row[pkey]['project']]
                self.assertItemsEqual(['CMSSW_5_3_6', 'CMSSW_5_3_7', 'CMSSW_5_3_8'], releases)
            elif row[pkey]['name'] == 'slc7_ppc64le_gcc530':
                releases = [item['label'] for item in row[pkey]['project']]
                self.assertItemsEqual(['CMSSW_8_1_0_pre15'], releases)
            elif row[pkey]['name'] == 'slc6_amd64_gcc620':
                releases = [item['label'] for item in row[pkey]['project']]
                self.assertItemsEqual(['CMSSW_9_0_0_pre3', 'CMSSW_9_0_0_pre4'], releases)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_FixedDelay_t_

Fixed Delay splitting test.
"""
from __future__ import division

from builtins import range

import unittest

from WMCore.DataStructs.File import File
from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.Job import Job
from WMCore.DataStructs.Subscription import Subscription
from WMCore.DataStructs.Workflow import Workflow
from WMCore.DataStructs.Run import Run
import time
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID

class FixedDelayTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Create two subscriptions: One that contains a single file and one that
        contains multiple files.
        """
        self.multipleFileFileset = Fileset(name = "TestFileset1")
        for i in range(10):
            newFile = File(makeUUID(), size = 1000, events = 100)
            newFile.addRun(Run(i, *[45+i]))
            self.multipleFileFileset.addFile(newFile)

        self.singleFileFileset = Fileset(name = "TestFileset2")
        newFile = File("/some/file/name", size = 1000, events = 100)
        newFile.addRun(Run(1, *[45]))
        self.singleFileFileset.addFile(newFile)

        self.multipleFileLumiset = Fileset(name = "TestFileset3")
        for i in range(10):
            newFile = File(makeUUID(), size = 1000, events = 100)
            newFile.addRun(Run(1, *[45 + i // 3]))
            self.multipleFileLumiset.addFile(newFile)

        self.singleLumiFileset = Fileset(name = "TestFileset4")
        for i in range(10):
            newFile = File(makeUUID(), size = 1000, events = 100)
            newFile.addRun(Run(1, *[45]))
            self.singleLumiFileset.addFile(newFile)


        testWorkflow = Workflow()
        self.multipleFileSubscription  = Subscription(fileset = self.multipleFileFileset,
                                                      workflow = testWorkflow,
                                                      split_algo = "FixedDelay",
                                                      type = "Processing")
        self.singleFileSubscription    = Subscription(fileset = self.singleFileFileset,
                                                      workflow = testWorkflow,
                                                      split_algo = "FixedDelay",
                                                      type = "Processing")
        self.multipleLumiSubscription  = Subscription(fileset = self.multipleFileLumiset,
                                                      workflow = testWorkflow,
                                                      split_algo = "FixedDelay",
                                                      type = "Processing")
        self.singleLumiSubscription    = Subscription(fileset = self.singleLumiFileset,
                                                      workflow = testWorkflow,
                                                      split_algo = "FixedDelay",
                                                      type = "Processing")


        return

    def tearDown(self):
        """
        _tearDown_

        Nothing to do...
        """
        pass

    def testNone(self):
        """
        _testNone_

        Since the time hasn'tpassed, we shouldn't get any jobs back.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)
        jobGroups = jobFactory(trigger_time = int(time.time()) * 2)
        self.assertEqual(jobGroups, [], "Should have returned a null set")

        jobFactory = splitter(self.multipleFileSubscription)
        jobGroups = jobFactory(trigger_time = int(time.time()) * 2)
        self.assertEqual(jobGroups, [], "Should have returned a null set")

        jobFactory = splitter(self.multipleLumiSubscription)
        jobGroups = jobFactory(trigger_time = int(time.time()) * 2)
        self.assertEqual(jobGroups, [], "Should have returned a null set")

        jobFactory = splitter(self.singleLumiSubscription)
        jobGroups = jobFactory(trigger_time = int(time.time()) * 2)
        self.assertEqual(jobGroups, [], "Should have returned a null set")

        return

    def testClosed(self):
        """
        _testClosed_
        since the subscriptions are closed and none of the files ahve been
        acquired, all of the files should show up
        """
        splitter = SplitterFactory()
        self.singleFileSubscription.getFileset().markOpen(False)
        jobFactory = splitter(self.singleFileSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        assert len(jobGroups) == 1, \
               "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 1, \
               "ERROR: JobFactory didn't create a single job."

        job = jobGroups[0].jobs.pop()

        assert job.getFiles(type = "lfn") == ["/some/file/name"], \
               "ERROR: Job contains unknown files."

        self.multipleFileSubscription.getFileset().markOpen(False)
        jobFactory = splitter(self.multipleFileSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs),1)
        myfiles = jobGroups[0].jobs[0].getFiles()
        self.assertEqual(len(myfiles), 10)

        self.multipleLumiSubscription.getFileset().markOpen(False)
        jobFactory = splitter(self.multipleLumiSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs),1)
        myfiles = jobGroups[0].jobs[0].getFiles()
        self.assertEqual(len(myfiles), 10)
        #self.assertEqual(jobGroups, [], "Should have returned a null set")

        self.singleLumiSubscription.getFileset().markOpen(False)
        jobFactory = splitter(self.singleLumiSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        assert len(jobGroups) == 1, \
               "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 1, \
               "ERROR: JobFactory didn't create a single job."
        myfiles = jobGroups[0].jobs[0].getFiles()
        self.assertEqual(len(myfiles), 10)


    def testAllAcquired(self):
        """
        _testAllAcquired_
        should all return no job groups
        """
        splitter = SplitterFactory()
        self.singleFileSubscription.acquireFiles(
                           self.singleFileSubscription.availableFiles())
        jobFactory = splitter(self.singleFileSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(jobGroups, [], "Should have returned a null set: %s" % jobGroups)

        self.multipleFileSubscription.acquireFiles(
                           self.multipleFileSubscription.availableFiles())
        jobFactory = splitter(self.multipleFileSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(jobGroups, [], "Should have returned a null set")

        self.multipleLumiSubscription.acquireFiles(
                           self.multipleLumiSubscription.availableFiles())
        jobFactory = splitter(self.multipleLumiSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(jobGroups, [], "Should have returned a null set")

        self.singleLumiSubscription.acquireFiles(
                           self.singleLumiSubscription.availableFiles())
        jobFactory = splitter(self.singleLumiSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(jobGroups, [], "Should have returned a null set")

    def testClosedSomeAcquired(self):
        """
        _testClosedSomeAcquired_
        since the subscriptions are closed and none of the files ahve been
        acquired, all of the files should show up
        """
        splitter = SplitterFactory()
        self.multipleFileSubscription.getFileset().markOpen(False)
        self.singleFileSubscription.acquireFiles(
                           [self.singleFileSubscription.availableFiles().pop()])
        jobFactory = splitter(self.singleFileSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(jobGroups, [], "Should have returned a null set")

        self.multipleFileSubscription.getFileset().markOpen(False)
        self.multipleFileSubscription.acquireFiles(
                           [self.multipleFileSubscription.availableFiles().pop()])
        jobFactory = splitter(self.multipleFileSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(len(jobGroups),1, "Should have gotten one jobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1, \
               "JobFactory should have made one job")
        myfiles = jobGroups[0].jobs[0].getFiles()
        self.assertEqual(len(myfiles), 9, \
                "JobFactory should have provides us with 9 files")

        self.multipleLumiSubscription.getFileset().markOpen(False)
        self.multipleLumiSubscription.acquireFiles(
                           [self.multipleLumiSubscription.availableFiles().pop()])
        jobFactory = splitter(self.multipleLumiSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(len(jobGroups),1, "Should have gotten one jobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1, \
               "JobFactory should have made one job")
        myfiles = jobGroups[0].jobs[0].getFiles()
        self.assertEqual(len(myfiles), 9, \
                "JobFactory should have provides us with 9 files")

        self.singleLumiSubscription.getFileset().markOpen(False)
        self.singleLumiSubscription.acquireFiles(
                           [self.singleLumiSubscription.availableFiles().pop()])
        jobFactory = splitter(self.singleLumiSubscription)
        jobGroups = jobFactory(trigger_time = 1)
        self.assertEqual(len(jobGroups),1, "Should have gotten one jobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1, \
               "JobFactory should have made one job")
        myfiles = jobGroups[0].jobs[0].getFiles()
        self.assertEqual(len(myfiles), 9, \
                "JobFactory should have provides us with 9 files")

        self.assertEqual(len(myfiles), 9)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_SplitFileBased_t_

Unit tests for the split file job splitting algorithm.
"""

import unittest
import os
import threading

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow

from WMCore.DataStructs.Run import Run

from WMCore.DAOFactory import DAOFactory
from WMCore.WMFactory import WMFactory
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID
from WMQuality.TestInit import TestInit

class SplitFileBasedTest(unittest.TestCase):
    """
    _SplitFileBasedTest_

    Unit tests for the split file job splitting algorithm.
    """

    def setUp(self):
        """
        _setUp_

        Create database connection and load up the WMBS schema.
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules = ["WMCore.WMBS"],
                                useDefault = False)

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package = "WMCore.WMBS",
                                     logger = myThread.logger,
                                     dbinterface = myThread.dbi)

        return

    def tearDown(self):
        """
        _tearDown_

        Clear out WMBS.
        """
        self.testInit.clearDatabase()
        return

    def stuffWMBS(self):
        """
        _stuffWMBS_

        Insert some dummy jobs, jobgroups, filesets, files and subscriptions
        into WMBS to test job creation.  Three completed job groups each
        containing several files are injected.  Another incomplete job group is
        also injected.  Also files are added to the "Mergeable" subscription as
        well as to the output fileset for their jobgroups.
        """
        locationAction = self.daoFactory(classname = "Locations.New")
        changeStateDAO = self.daoFactory(classname = "Jobs.ChangeState")

        locationAction.execute("site1", pnn = "T2_CH_CERN")

        self.mergeFileset = Fileset(name = "mergeFileset")
        self.mergeFileset.create()
        self.bogusFileset = Fileset(name = "bogusFileset")
        self.bogusFileset.create()

        self.mergeMergedFileset = Fileset(name = "mergeMergedFileset")
        self.mergeMergedFileset.create()
        self.bogusMergedFileset = Fileset(name = "bogusMergedFileset")
        self.bogusMergedFileset.create()

        mergeWorkflow = Workflow(name = "mergeWorkflow", spec = "bunk2",
                                 owner = "Steve", task="Test")
        mergeWorkflow.create()

        self.mergeSubscription = Subscription(fileset = self.mergeFileset,
                                              workflow = mergeWorkflow,
                                              split_algo = "SplitFileBased")
        self.mergeSubscription.create()
        self.bogusSubscription = Subscription(fileset = self.bogusFileset,
                                              workflow = mergeWorkflow,
                                              split_algo = "SplitFileBased")

        inputFileset = Fileset(name = "inputFileset")
        inputFileset.create()

        inputWorkflow = Workflow(name = "inputWorkflow", spec = "input",
                                owner = "Steve", task = "Test")
        inputWorkflow.create()
        inputWorkflow.addOutput("someOutput", self.mergeFileset,
                                self.mergeMergedFileset)
        inputWorkflow.addOutput("someOutput2", self.bogusFileset,
                                self.bogusMergedFileset)

        inputSubscription = Subscription(fileset = inputFileset,
                                        workflow = inputWorkflow)
        inputSubscription.create()

        parentFile1 = File(lfn = "parentFile1")
        parentFile1.create()
        parentFile2 = File(lfn = "parentFile2")
        parentFile2.create()
        parentFile3 = File(lfn = "parentFile3")
        parentFile3.create()
        parentFile4 = File(lfn = "parentFile4")
        parentFile4.create()

        jobGroup1 = JobGroup(subscription = inputSubscription)
        jobGroup1.create()
        jobGroup2 = JobGroup(subscription = inputSubscription)
        jobGroup2.create()

        testJob1 = Job()
        testJob1.addFile(parentFile1)
        testJob1.create(jobGroup1)
        testJob1["state"] = "cleanout"
        testJob1["oldstate"] = "new"
        testJob1["couch_record"] = "somejive"
        testJob1["retry_count"] = 0
        testJob1["outcome"] = "success"
        testJob1.save()
        changeStateDAO.execute([testJob1])

        testJob2 = Job()
        testJob2.addFile(parentFile2)
        testJob2.create(jobGroup1)
        testJob2["state"] = "cleanout"
        testJob2["oldstate"] = "new"
        testJob2["couch_record"] = "somejive"
        testJob2["retry_count"] = 0
        testJob2["outcome"] = "success"
        testJob2.save()
        changeStateDAO.execute([testJob2])

        testJob3 = Job()
        testJob3.addFile(parentFile3)
        testJob3.create(jobGroup2)
        testJob3["state"] = "cleanout"
        testJob3["oldstate"] = "new"
        testJob3["couch_record"] = "somejive"
        testJob3["retry_count"] = 0
        testJob3["outcome"] = "success"
        testJob3.save()
        changeStateDAO.execute([testJob3])

        testJob4 = Job()
        testJob4.addFile(parentFile4)
        testJob4.create(jobGroup2)
        testJob4["state"] = "cleanout"
        testJob4["oldstate"] = "new"
        testJob4["couch_record"] = "somejive"
        testJob4["retry_count"] = 0
        testJob4["outcome"] = "failure"
        testJob4.save()
        changeStateDAO.execute([testJob4])

        # We'll simulate a failed split by event job that the merger should
        # ignore.
        parentFile5 = File(lfn = "parentFile5")
        parentFile5.create()

        testJob5 = Job()
        testJob5.addFile(parentFile5)
        testJob5.create(jobGroup2)
        testJob5["state"] = "cleanout"
        testJob5["oldstate"] = "new"
        testJob5["couch_record"] = "somejive"
        testJob5["retry_count"] = 0
        testJob5["outcome"] = "success"
        testJob5.save()
        changeStateDAO.execute([testJob5])

        testJob6 = Job()
        testJob6.addFile(parentFile5)
        testJob6.create(jobGroup2)
        testJob6["state"] = "cleanout"
        testJob6["oldstate"] = "new"
        testJob6["couch_record"] = "somejive"
        testJob6["retry_count"] = 0
        testJob6["outcome"] = "failure"
        testJob6.save()
        changeStateDAO.execute([testJob6])

        badFile1 = File(lfn = "badFile1", size = 10241024, events = 10241024, first_event = 0, locations = set(["T2_CH_CERN"]))
        badFile1.addRun(Run(1, *[45]))
        badFile1.create()
        badFile1.addParent(parentFile5["lfn"])

        file1 = File(lfn = "file1", size = 1024, events = 1024, first_event = 0, locations = set(["T2_CH_CERN"]))
        file1.addRun(Run(1, *[45]))
        file1.create()
        file1.addParent(parentFile1["lfn"])
        file2 = File(lfn = "file2", size = 1024, events = 1024, first_event = 1024, locations = set(["T2_CH_CERN"]))
        file2.addRun(Run(1, *[45]))
        file2.create()
        file2.addParent(parentFile1["lfn"])
        file3 = File(lfn = "file3", size = 1024, events = 1024, first_event = 2048, locations = set(["T2_CH_CERN"]))
        file3.addRun(Run(1, *[45]))
        file3.create()
        file3.addParent(parentFile1["lfn"])
        file4 = File(lfn = "file4", size = 1024, events = 1024, first_event = 3072, locations = set(["T2_CH_CERN"]))
        file4.addRun(Run(1, *[45]))
        file4.create()
        file4.addParent(parentFile1["lfn"])

        fileA = File(lfn = "fileA", size = 1024, events = 1024, first_event = 0, locations = set(["T2_CH_CERN"]))
        fileA.addRun(Run(1, *[46]))
        fileA.create()
        fileA.addParent(parentFile2["lfn"])
        fileB = File(lfn = "fileB", size = 1024, events = 1024, first_event = 1024, locations = set(["T2_CH_CERN"]))
        fileB.addRun(Run(1, *[46]))
        fileB.create()
        fileB.addParent(parentFile2["lfn"])
        fileC = File(lfn = "fileC", size = 1024, events = 1024, first_event = 2048, locations = set(["T2_CH_CERN"]))
        fileC.addRun(Run(1, *[46]))
        fileC.create()
        fileC.addParent(parentFile2["lfn"])

        fileI = File(lfn = "fileI", size = 1024, events = 1024, first_event = 0, locations = set(["T2_CH_CERN"]))
        fileI.addRun(Run(2, *[46]))
        fileI.create()
        fileI.addParent(parentFile3["lfn"])
        fileII = File(lfn = "fileII", size = 1024, events = 1024, first_event = 1024, locations = set(["T2_CH_CERN"]))
        fileII.addRun(Run(2, *[46]))
        fileII.create()
        fileII.addParent(parentFile3["lfn"])
        fileIII = File(lfn = "fileIII", size = 1024, events = 1024, first_event = 2048, locations = set(["T2_CH_CERN"]))
        fileIII.addRun(Run(2, *[46]))
        fileIII.create()
        fileIII.addParent(parentFile3["lfn"])
        fileIV = File(lfn = "fileIV", size = 1024, events = 1024, first_event = 3072, locations = set(["T2_CH_CERN"]))
        fileIV.addRun(Run(2, *[46]))
        fileIV.create()
        fileIV.addParent(parentFile3["lfn"])

        fileX = File(lfn = "badFileA", size = 1024, events = 1024, first_event = 0, locations = set(["T2_CH_CERN"]))
        fileX.addRun(Run(1, *[47]))
        fileX.create()
        fileX.addParent(parentFile4["lfn"])
        fileY = File(lfn = "badFileB", size = 1024, events = 1024, first_event = 1024, locations = set(["T2_CH_CERN"]))
        fileY.addRun(Run(1, *[47]))
        fileY.create()
        fileY.addParent(parentFile4["lfn"])
        fileZ = File(lfn = "badFileC", size = 1024, events = 1024, first_event = 2048, locations = set(["T2_CH_CERN"]))
        fileZ.addRun(Run(1, *[47]))
        fileZ.create()
        fileZ.addParent(parentFile4["lfn"])

        jobGroup1.output.addFile(file1)
        jobGroup1.output.addFile(file2)
        jobGroup1.output.addFile(file3)
        jobGroup1.output.addFile(file4)
        jobGroup1.output.addFile(fileA)
        jobGroup1.output.addFile(fileB)
        jobGroup1.output.addFile(fileC)
        jobGroup1.output.commit()

        jobGroup2.output.addFile(fileI)
        jobGroup2.output.addFile(fileII)
        jobGroup2.output.addFile(fileIII)
        jobGroup2.output.addFile(fileIV)
        jobGroup2.output.addFile(fileX)
        jobGroup2.output.addFile(fileY)
        jobGroup2.output.addFile(fileZ)
        jobGroup2.output.addFile(badFile1)
        jobGroup2.output.commit()

        for file in [file1, file2, file3, file4, fileA, fileB, fileC, fileI,
                     fileII, fileIII, fileIV, fileX, fileY, fileZ, badFile1]:
            self.mergeFileset.addFile(file)
            self.bogusFileset.addFile(file)

        self.mergeFileset.commit()
        self.bogusFileset.commit()

        return

    def testSplitAlgo(self):
        """
        _testSplitAlgo_

        Run the SplitFileBased splitting algorithm over the data created in
        the merge subscription.  This should produce three job groups each
        containing one job.  The files in the job should be ordered correctly.
        """
        self.stuffWMBS()

        splitter = SplitterFactory()
        jobFactory = splitter(package = "WMCore.WMBS",
                              subscription = self.mergeSubscription)

        result = jobFactory()

        assert len(result) == 3, \
               "ERROR: Wrong number of job groups returned: %s" % len(result)

        for jobGroup in result:
            assert len(jobGroup.jobs) == 1, \
                   "ERROR: One job should be in a job group."

        goldenFilesA = ["file1", "file2", "file3", "file4"]
        goldenFilesB = ["fileA", "fileB", "fileC"]
        goldenFilesC = ["fileI", "fileII", "fileIII", "fileIV"]

        for jobGroup in result:
            jobFiles = jobGroup.jobs.pop().getFiles()

            if jobFiles[0]["lfn"] in goldenFilesA:
                goldenFiles = goldenFilesA
            elif jobFiles[0]["lfn"] in goldenFilesB:
                goldenFiles = goldenFilesB
            else:
                goldenFiles = goldenFilesC

            currentRun = 0
            currentLumi = 0
            currentEvent = 0
            for file in jobFiles:
                file.loadData()
                assert file["lfn"] in goldenFiles, \
                       "Error: Unknown file in merge jobs."
                assert len(file["locations"]) == 1, \
                       "Error: Wrong number of file locations."
                assert "T2_CH_CERN" in file["locations"], \
                       "Error: File is missing a location."

                goldenFiles.remove(file["lfn"])

            fileRun = list(file["runs"])[0].run
            fileLumi = min(list(file["runs"])[0])
            fileEvent = file["first_event"]

            if currentRun == 0:
                currentRun = fileRun
                currentLumi = fileLumi
                currentEvent = fileEvent
                continue

            assert fileRun >= currentRun, \
                   "ERROR: Files not sorted by run."

            if fileRun == currentRun:
                assert fileLumi >= currentLumi, \
                       "ERROR: Files not ordered by lumi"

                if fileLumi == currentLumi:
                    assert fileEvent > currentEvent, \
                           "ERROR: Files not ordered by first event"

            currentRun = fileRun
            currentLumi = fileLumi
            currentEvent = fileEvent

        assert len(goldenFilesA) == 0 and len(goldenFilesB) == 0 and \
               len(goldenFilesC) == 0, \
               "ERROR: Files missing from merge jobs."

        return

if __name__ == '__main__':
    unittest.main()



"""
_ProcessPoolTestWorker_

"""

from WMCore.ProcessPool.ProcessPool import ProcessPoolWorker

class ProcessPoolTestWorker(ProcessPoolWorker):

    def __call__(self, input):
        """
        __call__

        """

        return input



from WMCore.REST.Main import ProfiledApp
from WMCore.REST.Main import Logger
from WMCore.REST.Main import RESTMain
from WMCore.REST.Main import RESTDaemon
from WMCore.REST.Test import fake_authz_key_file
from WMCore.Configuration import Configuration
from cherrypy import Application
import os

def dummy():
    return ""

import os

authz_key = fake_authz_key_file()
cfg = Configuration()
main = cfg.section_('main')
main.application = 'test'
main.silent = True
main.index = 'top'
main.authz_defaults = { 'role': None, 'group': None, 'site': None }
main.section_('tools').section_('cms_auth').key_file = authz_key.name
app = cfg.section_('test')
app.admin = 'dada@example.org'
app.description = app.title = 'Test'
views = cfg.section_('views')
top = views.section_('top')
top.object = os.path.abspath(__file__).rsplit("/", 1)[-1].split(".")[0] + ".Test"

ProfiledApp(Application(dummy, "/", {"/": {}}), ".")
Logger()
RESTMain(cfg, ".")
RESTDaemon(cfg, ".")



#!/usr/bin/env python
# encoding: utf-8
"""
User_t.py

Created by Dave Evans on 2010-07-29.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

import unittest
import os

from WMCore.Database.CMSCouch import Document
from WMCore.GroupUser.User import User
from WMCore.GroupUser.Group import Group


class User_t(unittest.TestCase):
    def setUp(self):
        self.database = "groupuser"
        self.url = os.getenv("COUCHURL", "http://127.0.0.1:5984")

    def testA(self):
        """instantiate & jsonise"""

        u1 = User(name = "evansde77")
        g1 = Group(name = "DMWM", administrators = ["evansde77", "drsm79"])
        g1.setCouch(self.url, self.database)
        g1.connect()
        u1. setGroup(g1)

        u1.create()

        u2 = User(name = "evansde77")
        u2.setCouch(self.url, self.database)
        u2.get()

        u1.drop()
        g1.drop()


    def testB(self):
        """test owning some sample documents"""

        u1 = User(name = "evansde77")
        g1 = Group(name = "DMWM", administrators = ["evansde77", "drsm79"])
        g1.setCouch(self.url, self.database)
        g1.connect()
        u1.setGroup(g1)
        u1.create()

        doc1 = Document()
        doc1['test-data'] = {"key1" : "value1"}
        doc2 = Document()
        doc2['test-data'] = {"key2" : "value2"}
        id1 = g1.couch.commitOne(doc1)[0]
        id2 = g1.couch.commitOne(doc2)[0]
        doc1['_id'] = id1[u'id']
        doc1['_rev'] = id1[u'rev']
        doc2['_id'] = id2[u'id']
        doc2['_rev'] = id2[u'rev']

        u1.ownThis(doc1)
        u1.ownThis(doc2)

        self.assertTrue("owner" in doc1)
        self.assertTrue("owner" in doc2)
        self.assertTrue('user' in doc1['owner'])
        self.assertTrue('group' in doc1['owner'])
        self.assertTrue(doc1['owner']['user'] == u1['name'])
        self.assertTrue(doc1['owner']['group'] == u1['group'])
        self.assertTrue('user' in doc2['owner'])
        self.assertTrue('group' in doc2['owner'])
        self.assertTrue(doc2['owner']['user'] == u1['name'])
        self.assertTrue(doc2['owner']['group'] == u1['group'])


        g1.couch.delete_doc(id1[u'id'])
        g1.couch.delete_doc(id2[u'id'])
        u1.drop()
        g1.drop()


if __name__ == '__main__':
    unittest.main()



"""
MicroService configuration file.
"""
from __future__ import division

import os
import socket
from WMCore.Configuration import Configuration

# globals
HOST = socket.gethostname().lower()
ROOTDIR = os.getenv('MS_STATIC_ROOT', os.getcwd())
config = Configuration()

main = config.section_("main")
srv = main.section_("server")
srv.thread_pool = 30
main.application = "microservice"
main.port = 8833  # main application port it listens on
main.index = 'ui' # Configuration requires index attribute
# Security configuration
#main.authz_defaults = {"role": None, "group": None, "site": None}
#sec = main.section_("tools").section_("cms_auth")
#sec.key_file = "%s/auth/wmcore-auth/header-auth-key" % ROOTDIR

# this is where the application will be mounted, where the REST API
# is reachable and this features in CMS web frontend rewrite rules
app = config.section_(main.application)
app.admin = "cms-service-webtools@cern.ch"
app.description = "CMS data operations MicroService"
app.title = "CMS MicroService"

# define different views for our application
views = config.section_("views")
# web UI interface
ui = views.section_('ui')
ui.object = 'WMCore.MicroService.WebGui.FrontPage.FrontPage'
ui.static = ROOTDIR

# REST interface
data = views.section_('data')
data.object = 'WMCore.MicroService.Service.RestApiHub.RestApiHub'
data.manager = 'WMCore.MicroService.MSManager.MSManager'



"""
_WorkQueueTestCase_

Unit tests for the WMBS File class.
"""

from __future__ import print_function

import time
import unittest

from Utils.PythonVersion import PY3

from WMCore.Agent.HeartbeatAPI import HeartbeatAPI
from WMQuality.TestInit import TestInit


class HeartbeatTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Setup the database and logging connection.  Try to create all of the
        Heartbeat tables.  Also add some dummy locations.
        """

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()  # logLevel = logging.SQLDEBUG
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.Agent.Database"],
                                useDefault=False)

        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def tearDown(self):
        """
        _tearDown_

        Drop all the Heartbeat tables.
        """

        self.testInit.clearDatabase()

    def testAddComponent(self):
        """
        _testAddComponent_

        Test creation of components and worker threads as well as the
        get heartbeat DAOs
        """
        comp1 = HeartbeatAPI("testComponent1", pollInterval=60, heartbeatTimeout=600)
        comp1.registerComponent()
        self.assertEqual(comp1.getHeartbeatInfo(), [])  # no worker thread yet

        comp1.registerWorker("testWorker1")
        self.assertEqual(len(comp1.getHeartbeatInfo()), 1)

        comp1.registerWorker("testWorker2")
        self.assertEqual(len(comp1.getHeartbeatInfo()), 2)

        comp2 = HeartbeatAPI("testComponent2", pollInterval=30, heartbeatTimeout=300)
        comp2.registerComponent()
        self.assertEqual(comp2.getHeartbeatInfo(), [])  # no worker thread yet
        self.assertEqual(len(comp2.getAllHeartbeatInfo()), 2)

        comp2.registerWorker("testWorker21")
        self.assertEqual(len(comp2.getHeartbeatInfo()), 1)
        self.assertEqual(len(comp2.getAllHeartbeatInfo()), 3)

        comp1.updateWorkerHeartbeat("testWorker1", "Running")
        comp1.updateWorkerHeartbeat("testWorker2", "Running")
        comp2.updateWorkerHeartbeat("testWorker21", "Running")
        self.assertEqual(len(comp1.getAllHeartbeatInfo()), 3)
        self.assertEqual(len(comp2.getAllHeartbeatInfo()), 3)

        comp1Res = comp1.getHeartbeatInfo()
        comp2Res = comp2.getHeartbeatInfo()
        self.assertEqual(len(comp1Res), 2)
        self.assertEqual(len(comp2Res), 1)

        self.assertItemsEqual([item["name"] for item in comp1Res], ["testComponent1", "testComponent1"])
        self.assertItemsEqual([item["worker_name"] for item in comp1Res], ["testWorker1", "testWorker2"])
        self.assertItemsEqual([item["state"] for item in comp1Res], ["Running", "Running"])
        self.assertItemsEqual([item["poll_interval"] for item in comp1Res], [60, 60])
        self.assertItemsEqual([item["update_threshold"] for item in comp1Res], [600, 600])

        self.assertItemsEqual([item["name"] for item in comp2Res], ["testComponent2"])
        self.assertItemsEqual([item["worker_name"] for item in comp2Res], ["testWorker21"])
        self.assertItemsEqual([item["state"] for item in comp2Res], ["Running"])
        self.assertItemsEqual([item["poll_interval"] for item in comp2Res], [30])
        self.assertItemsEqual([item["update_threshold"] for item in comp2Res], [300])

    def testUpdateWorkers(self):
        """
        _testUpdateWorkers_

        Create a couple of components and workers and test the update methods
        """
        comp1 = HeartbeatAPI("testComponent1", pollInterval=60, heartbeatTimeout=600)
        comp1.registerComponent()
        comp1.registerWorker("testWorker1")
        comp1.registerWorker("testWorker2")

        comp2 = HeartbeatAPI("testComponent2", pollInterval=30, heartbeatTimeout=300)
        comp2.registerComponent()
        comp2.registerWorker("testWorker21")

        comp1.updateWorkerCycle("testWorker1", 1.001, None)
        comp2.updateWorkerCycle("testWorker21", 1234.1, 100)
        hb1 = comp1.getHeartbeatInfo()
        hb2 = comp2.getHeartbeatInfo()

        for worker in hb1:
            if worker['worker_name'] == 'testWorker1':
                self.assertTrue(worker["cycle_time"] > 1.0)
            else:
                self.assertEqual(worker["cycle_time"], 0)
        self.assertItemsEqual([item["outcome"] for item in hb1], [None, None])
        self.assertItemsEqual([item["error_message"] for item in hb1], [None, None])

        self.assertEqual(round(hb2[0]["cycle_time"], 1), 1234.1)
        self.assertEqual(hb2[0]["outcome"], '100')
        self.assertEqual(hb2[0]["error_message"], None)

        # time to update workers with an error
        comp1.updateWorkerError("testWorker2", "BAD JOB!!!")
        hb1 = comp1.getHeartbeatInfo()
        for worker in hb1:
            if worker['worker_name'] == 'testWorker2':
                self.assertTrue(worker["last_error"] > int(time.time() - 10))
                self.assertEqual(worker["state"], "Error")
                self.assertEqual(worker["error_message"], "BAD JOB!!!")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_DeleteCheckFile_

MySQL implementation of DeleteCheckFile

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class DeleteCheck(DBFormatter):
    sql = """DELETE FROM wmbs_file_details WHERE id = :id AND
           NOT EXISTS (SELECT fileset FROM wmbs_fileset_files WHERE fileid = :id
           AND fileset != :fileset)"""


    def execute(self, file = None, fileset = None, conn = None, transaction = False):
        if isinstance(file, list):
            if len(file) < 1:
                return
            binds = []
            for entry in file:
                binds.append({'id': entry, 'fileset': fileset})
        else:
            binds = {'id': file, 'fileset': fileset}
        self.dbi.processData(self.sql, binds,
                         conn = conn, transaction = transaction)
        return True #or raise



#/usr/bin/env python2.4
"""
_Destroy_

"""




from WMCore.ThreadPool.MySQL.Destroy import Destroy as MySQLDestroy
from WMCore.ThreadPool.Oracle.Create import Create

class Destroy(MySQLDestroy):
    def __init__(self, logger = None, dbi = None):
        """
        _init_

        Call the base class's constructor and add all necessary tables for
        deletion,
        """
        MySQLDestroy.__init__(self, logger, dbi)

        j=50
        for i in Create.sequence_tables:
            seqname = i
            self.delete["%s%s" % (j, seqname)] = \
                           "DROP SEQUENCE %s"  % seqname


        #for i in Create.trigger_tables:
        #    seqname = i
        #    self.create["%s%s" % (j, seqname)] = \
        #                   "DROP TRIGGER %s"  % seqname



#!/usr/bin/env python
"""
_DBSReader_t_

Unit test for the DBS helper class.
"""

import unittest

from RestClient.ErrorHandling.RestClientExceptions import HTTPError
from nose.plugins.attrib import attr

from Utils.PythonVersion import PY3

from WMCore.Services.DBS.DBS3Reader import getDataTiers, DBS3Reader as DBSReader
from WMCore.Services.DBS.DBSErrors import DBSReaderError
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase

# A small dataset that should always exist
BAD_DATASET = '/HighPileUp/Run2011A-v1/RAW-BLAH'
DATASET = '/HighPileUp/Run2011A-v1/RAW'
BLOCK = '/HighPileUp/Run2011A-v1/RAW#fabf118a-cbbf-11e0-80a9-003048caaace'
FILE = '/store/data/Run2011A/HighPileUp/RAW/v1/000/173/657/B293AF24-BFCB-E011-8F85-BCAEC5329701.root'
FILE2 = '/store/data/Run2011A/HighPileUp/RAW/v1/000/173/660/EE0B83F9-F3CB-E011-B996-BCAEC5329710.root'

# A RECO dataset that has parents (also small)
DATASET_WITH_PARENTS = '/Cosmics/ComissioningHI-PromptReco-v1/RECO'
BLOCK_WITH_PARENTS = DATASET_WITH_PARENTS + '#7020873e-0dcd-11e1-9b6c-003048caaace'
FILE1_WITH_PARENT = '/store/data/ComissioningHI/Cosmics/RECO/PromptReco-v1/000/180/841/368B76AA-4F09-E111-82CB-BCAEC5329721.root'
FILE2_WITH_PARENT = '/store/data/ComissioningHI/Cosmics/RECO/PromptReco-v1/000/180/852/D82D6996-1B0B-E111-AA40-003048CF94A6.root'

PARENT_DATASET = '/Cosmics/ComissioningHI-v1/RAW'
PARENT_BLOCK = PARENT_DATASET + '#929366bc-0c31-11e1-b764-003048caaace'
PARENT_FILE = '/store/data/ComissioningHI/Cosmics/RAW/v1/000/181/369/662EAD44-300C-E111-A709-BCAEC518FF62.root'


class DBSReaderTest(EmulatedUnitTestCase):
#class DBSReaderTest(unittest.TestCase):

    def setUp(self):
        """
        _setUp_

        Initialize the PhEDEx API to point at the test server.
        """

        self.endpoint = 'https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader'
        self.dbs = None
        super(DBSReaderTest, self).setUp()
        if PY3:
            self.assertItemsEqual = self.assertCountEqual
        return

    def tearDown(self):
        """
        _tearDown_

        """

        super(DBSReaderTest, self).tearDown()
        return

    def testListDatatiers(self):
        """
        listDatatiers returns all datatiers available
        """
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.listDatatiers()
        self.assertTrue('RAW' in results)
        self.assertTrue('GEN-SIM-RECO' in results)
        self.assertTrue('GEN-SIM' in results)
        self.assertFalse('RAW-ALAN' in results)
        return

    def testGetDataTiers(self):
        """
        Test the getDataTiers function
        """
        results = getDataTiers(self.endpoint)
        self.assertTrue('RAW' in results)
        self.assertTrue('GEN-SIM-RECO' in results)
        self.assertTrue('GEN-SIM' in results)
        self.assertFalse('RAW-ALAN' in results)
        # dbsUrl is mandatory
        with self.assertRaises(TypeError):
            _ = getDataTiers()
        return

    def testListPrimaryDatasets(self):
        """
        listPrimaryDatasets returns known primary datasets
        """
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.listPrimaryDatasets('Jet*')
        self.assertTrue('Jet' in results)
        self.assertTrue('JetMET' in results)
        self.assertTrue('JetMETTau' in results)
        self.assertFalse(self.dbs.listPrimaryDatasets('DoesntExist'))
        return

    def testMatchProcessedDatasets(self):
        """
        matchProcessedDatasets returns known processed datasets
        """
        self.dbs = DBSReader(self.endpoint)
        dataset = self.dbs.matchProcessedDatasets('Jet', 'RAW', 'Run2011A-v1')
        self.assertEqual(1, len(dataset))
        self.assertEqual(['/Jet/Run2011A-v1/RAW'], dataset[0]['PathList'])
        self.assertEqual('Run2011A-v1', dataset[0]['Name'])
        self.assertFalse(self.dbs.matchProcessedDatasets('Jet', 'RAW', 'Run2011A-v666'))

    def testlistRuns(self):
        """listRuns returns known runs"""
        self.dbs = DBSReader(self.endpoint)
        runs = self.dbs.listRuns(dataset=DATASET)
        self.assertEqual(46, len(runs))
        self.assertTrue(174074 in runs)
        runs = self.dbs.listRuns(block=BLOCK)
        self.assertEqual(1, len(runs))
        self.assertEqual([173657], runs)

    def testlistRunLumis(self):
        """listRunLumis returns known runs and lumicounts (None for DBS3)"""
        self.dbs = DBSReader(self.endpoint)
        runs = self.dbs.listRunLumis(dataset=DATASET)
        self.assertEqual(46, len(runs))
        self.assertTrue(173692 in runs)
        self.assertEqual(runs[173692], None)
        runs = self.dbs.listRunLumis(block=BLOCK)
        self.assertEqual(1, len(runs))
        self.assertTrue(173657 in runs)
        self.assertEqual(runs[173657], None)

    def testListProcessedDatasets(self):
        """listProcessedDatasets returns known processed datasets"""
        self.dbs = DBSReader(self.endpoint)
        datasets = self.dbs.listProcessedDatasets('Jet', 'RAW')
        self.assertTrue('Run2011A-v1' in datasets)
        self.assertTrue('Run2011B-v1' in datasets)
        self.assertFalse(self.dbs.listProcessedDatasets('blah', 'RAW'))
        # with the migration to Go DBS, it now raises an HTTPError exception like
        # Function:dbs.validator.Check Message:unable to match 'data_tier_name' value 'blah' Error: invalid parameter(s)"
        with self.assertRaises(HTTPError):
            self.dbs.listProcessedDatasets('Jet', 'blah')

    def testlistDatasetFiles(self):
        """listDatasetFiles returns files in dataset"""
        self.dbs = DBSReader(self.endpoint)
        files = self.dbs.listDatasetFiles(DATASET)
        self.assertEqual(49, len(files))
        self.assertTrue(FILE in files)

    def testlistDatasetFileDetails(self):
        """testlistDatasetFilesDetails returns lumis, events, and parents of a dataset"""
        TESTFILE = '/store/data/Run2011A/HighPileUp/RAW/v1/000/173/658/56484BAB-CBCB-E011-AF00-BCAEC518FF56.root'
        self.dbs = DBSReader(self.endpoint)
        details = self.dbs.listDatasetFileDetails(DATASET)
        self.assertEqual(len(details), 49)
        self.assertTrue(TESTFILE in details)
        self.assertEqual(details[TESTFILE]['NumberOfEvents'], 545)
        self.assertEqual(details[TESTFILE]['file_size'], 286021145)
        self.assertEqual(details[TESTFILE]['BlockName'],
                         '/HighPileUp/Run2011A-v1/RAW#dd6e0796-cbcc-11e0-80a9-003048caaace')
        self.assertEqual(details[TESTFILE]['Md5'], 'NOTSET')
        self.assertEqual(details[TESTFILE]['md5'], 'NOTSET')
        self.assertEqual(details[TESTFILE]['Adler32'], 'a41a1446')
        self.assertEqual(details[TESTFILE]['adler32'], 'a41a1446')
        self.assertEqual(details[TESTFILE]['Checksum'], '22218315')
        self.assertEqual(details[TESTFILE]['check_sum'], '22218315')
        self.assertTrue(173658 in details[TESTFILE]['Lumis'])
        self.assertEqual(sorted(details[TESTFILE]['Lumis'][173658]),
                         [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26,
                          27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
                          50,
                          51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73,
                          74,
                          75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97,
                          98,
                          99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111])

    def testGetDBSSummaryInfo(self):
        """getDBSSummaryInfo returns summary of dataset and block"""
        self.dbs = DBSReader(self.endpoint)
        dataset = self.dbs.getDBSSummaryInfo(DATASET)
        self.assertEqual(dataset['path'], DATASET)
        self.assertEqual(dataset['block'], '')
        self.assertEqual(dataset['NumberOfEvents'], 22075)
        self.assertEqual(dataset['NumberOfBlocks'], 46)
        self.assertEqual(dataset['FileSize'], 4001680824)
        self.assertEqual(dataset['file_size'], 4001680824)
        self.assertEqual(dataset['NumberOfFiles'], 49)
        self.assertEqual(dataset['NumberOfLumis'], 7223)

        block = self.dbs.getDBSSummaryInfo(DATASET, BLOCK)
        self.assertEqual(block['path'], DATASET)
        self.assertEqual(block['block'], BLOCK)
        self.assertEqual(block['NumberOfEvents'], 377)
        self.assertEqual(block['NumberOfBlocks'], 1)
        self.assertEqual(block['FileSize'], 150780132)
        self.assertEqual(block['file_size'], 150780132)
        self.assertEqual(block['NumberOfFiles'], 2)
        self.assertEqual(block['NumberOfLumis'], 94)

        with self.assertRaises(DBSReaderError):
            self.dbs.getDBSSummaryInfo(BAD_DATASET)
        with self.assertRaises(DBSReaderError):
            self.dbs.getDBSSummaryInfo(DATASET, BLOCK + 'asas')

    def testListFileBlocks(self):
        """listFileBlocks returns block names in dataset"""
        self.dbs = DBSReader(self.endpoint)
        blocks = self.dbs.listFileBlocks(DATASET)
        self.assertTrue(BLOCK in blocks)
        # block is closed
        block = self.dbs.listFileBlocks(DATASET, blockName=BLOCK)[0]
        self.assertEqual(block, BLOCK)
        self.assertTrue(BLOCK in block)

    def testBlockExists(self):
        """blockExists returns existence of blocks"""
        self.dbs = DBSReader(self.endpoint)
        self.assertTrue(self.dbs.blockExists(BLOCK))
        self.assertFalse(self.dbs.blockExists(DATASET + '#somethingelse'))

    def testListFilesInBlock(self):
        """listFilesInBlock returns files in block"""
        self.dbs = DBSReader(self.endpoint)
        self.assertTrue(FILE in [x['LogicalFileName'] for x in self.dbs.listFilesInBlock(BLOCK)])
        self.assertRaises(DBSReaderError, self.dbs.listFilesInBlock, DATASET + '#blah')

    def testListFilesInBlockWithParents(self):
        """listFilesInBlockWithParents gets files with parents for a block"""
        self.dbs = DBSReader(self.endpoint)
        files = self.dbs.listFilesInBlockWithParents(
                '/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0')
        self.assertEqual(4, len(files))
        self.assertEqual('/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0',
                         files[0]['block_name'])
        self.assertEqual('/Cosmics/Commissioning2015-PromptReco-v1/RECO#004ac3ba-d09e-11e4-afad-001e67ac06a0',
                         files[0]['BlockName'])
        self.assertEqual(
                '/store/data/Commissioning2015/Cosmics/RAW/v1/000/238/545/00000/1043E89F-2DCF-E411-9CAE-02163E013751.root',
                files[0]['ParentList'][0]['LogicalFileName'])

        self.assertRaises(DBSReaderError, self.dbs.listFilesInBlockWithParents, BLOCK + 'asas')

    def testLfnsInBlock(self):
        """lfnsInBlock returns lfns in block"""
        self.dbs = DBSReader(self.endpoint)
        self.assertTrue(FILE in [x['logical_file_name'] for x in self.dbs.lfnsInBlock(BLOCK)])
        self.assertRaises(DBSReaderError, self.dbs.lfnsInBlock, BLOCK + 'asas')

    def testListFileBlockLocation(self):
        """listFileBlockLocation returns block location"""
        WRONG_BLOCK = BLOCK[:-4] + 'abcd'
        BLOCK2 = '/HighPileUp/Run2011A-v1/RAW#6021175e-cbfb-11e0-80a9-003048caaace'
        DBS_BLOCK = '/GenericTTbar/hernan-140317_231446_crab_JH_ASO_test_T2_ES_CIEMAT_5000_100_140318_0014-' + \
                    'ea0972193530f531086947d06eb0f121/USER#fb978442-a61b-413a-b4f4-526e6cdb142e'
        DBS_BLOCK2 = '/GenericTTbar/hernan-140317_231446_crab_JH_ASO_test_T2_ES_CIEMAT_5000_100_140318_0014-' + \
                     'ea0972193530f531086947d06eb0f121/USER#0b04d417-d734-4ef2-88b0-392c48254dab'
        self.dbs = DBSReader('https://cmsweb-prod.cern.ch/dbs/prod/phys03/DBSReader/')

        self.assertEqual(self.dbs.listFileBlockLocation(BLOCK), [])
        # This block is only found on DBS
        self.assertItemsEqual(self.dbs.listFileBlockLocation(DBS_BLOCK), [u'T2_ES_CIEMAT'])
        # doesn't raise on non-existant block
        self.assertEqual(self.dbs.listFileBlockLocation(WRONG_BLOCK), [])
        # test bulk call:
        ## two blocks in phedex
        self.assertEqual(2, len(self.dbs.listFileBlockLocation([BLOCK, BLOCK2])))
        ## one block in phedex one does not exist
        self.assertEqual(2, len(self.dbs.listFileBlockLocation([BLOCK, WRONG_BLOCK])))
        ## one in phedex one in dbs
        self.assertEqual(2, len(self.dbs.listFileBlockLocation([BLOCK, DBS_BLOCK])))
        ## two in dbs
        self.assertEqual(2, len(self.dbs.listFileBlockLocation([DBS_BLOCK, DBS_BLOCK2])))
        ## one in DBS and one does not exist
        self.assertEqual(2, len(self.dbs.listFileBlockLocation([DBS_BLOCK, WRONG_BLOCK])))

    def testGetFileBlock(self):
        """getFileBlock returns block"""
        self.dbs = DBSReader(self.endpoint)
        block = self.dbs.getFileBlock(BLOCK)
        self.assertEqual(len(block), 2)
        self.assertEqual(2, len(block['Files']))

        self.assertRaises(DBSReaderError, self.dbs.getFileBlock, BLOCK + 'asas')

    def testGetFileBlockWithParents(self):
        """getFileBlockWithParents returns block and parents"""
        self.dbs = DBSReader(self.endpoint)
        block = self.dbs.getFileBlockWithParents(BLOCK_WITH_PARENTS)
        self.assertEqual(len(block), 2)
        self.assertEqual(PARENT_FILE, block['Files'][0]['ParentList'][0]['LogicalFileName'])

        self.assertRaises(DBSReaderError, self.dbs.getFileBlockWithParents, BLOCK + 'asas')

    def testListBlockParents(self):
        """listBlockParents returns block parents"""
        self.dbs = DBSReader(self.endpoint)
        parents = self.dbs.listBlockParents(BLOCK_WITH_PARENTS)
        self.assertItemsEqual([PARENT_BLOCK], parents)

        self.assertFalse(self.dbs.listBlockParents(PARENT_BLOCK))

    def testBlockToDatasetPath(self):
        """blockToDatasetPath extracts path from block name"""
        self.dbs = DBSReader(self.endpoint)
        self.assertEqual(self.dbs.blockToDatasetPath(BLOCK), DATASET)
        self.assertRaises(DBSReaderError, self.dbs.blockToDatasetPath, BLOCK + 'asas')



    ### NEW UNIT TESTS ###
    def testCheckDBSServer(self):
        """Test the checkDBSServer method"""
        # Go and Python based servers return different data in different format
        # just check whether something is returned
        self.dbs = DBSReader(self.endpoint)
        self.assertTrue(self.dbs.checkDBSServer())

    def testCrossCheck(self):
        """Test the crossCheck method"""
        self.dbs = DBSReader(self.endpoint)
        listFiles = [FILE, FILE1_WITH_PARENT]
        results = self.dbs.crossCheck(DATASET, *listFiles)
        self.assertItemsEqual(results, [FILE])

    def testCrossCheckMissing(self):
        """Test the crossCheckMissing method"""
        self.dbs = DBSReader(self.endpoint)
        listFiles = [FILE, FILE1_WITH_PARENT]
        results = self.dbs.crossCheckMissing(DATASET, *listFiles)
        self.assertItemsEqual(results, [FILE1_WITH_PARENT])

    def testListDatasetLocation(self):
        """Test the listDatasetLocation method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.listDatasetLocation(DATASET)
        self.assertItemsEqual(results, [])

    def testCheckDatasetPath(self):
        """Test the checkDatasetPath method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.checkDatasetPath(DATASET)
        self.assertIsNone(results)

        with self.assertRaises(DBSReaderError):
            self.dbs.checkDatasetPath("")
        with self.assertRaises(DBSReaderError):
            self.dbs.checkDatasetPath(None)
        with self.assertRaises(DBSReaderError):
            self.dbs.checkDatasetPath(BAD_DATASET)

    def testCheckBlockName(self):
        """Test the checkBlockName method"""
        self.dbs = DBSReader(self.endpoint)
        self.assertIsNone(self.dbs.checkBlockName(DATASET))
        self.assertIsNone(self.dbs.checkBlockName(BLOCK))
        self.assertIsNone(self.dbs.checkBlockName("blah"))

        with self.assertRaises(DBSReaderError):
            self.assertIsNone(self.dbs.checkBlockName("*"))
        with self.assertRaises(DBSReaderError):
            self.assertIsNone(self.dbs.checkBlockName(""))
        with self.assertRaises(DBSReaderError):
            self.assertIsNone(self.dbs.checkBlockName(None))

    def testGetFileListByDataset(self):
        """Test the getFileListByDataset method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.getFileListByDataset(DATASET, detail=False)
        self.assertEqual(len(results), 49)
        self.assertEqual(len(results[0]), 1)
        self.assertTrue('logical_file_name' in results[0])
        self.assertTrue(results[0]['logical_file_name'].startswith("/store/data/Run2011A/HighPileUp/RAW/v1"))

        results = self.dbs.getFileListByDataset(DATASET, detail=True)
        self.assertEqual(len(results), 49)
        self.assertEqual(len(results[0]), 20)  # 20 keys in total
        subSetKeys = {'block_name', 'dataset', 'event_count', 'file_size', 'logical_file_name'}
        self.assertTrue(subSetKeys.issubset(results[0]))

    def testListDatasetParents(self):
        """Test the listDatasetParents method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.listDatasetParents(DATASET_WITH_PARENTS)
        self.assertEqual(results[0]['this_dataset'], DATASET_WITH_PARENTS)
        self.assertEqual(results[0]['parent_dataset'], PARENT_DATASET)

        results = self.dbs.listDatasetParents(PARENT_DATASET)
        self.assertEqual(results, [])

        with self.assertRaises(DBSReaderError):
            self.dbs.listDatasetParents(BLOCK)

    @attr('integration')  # too much data to mock
    def testGetParentFilesGivenParentDataset(self):
        """Test the getParentFilesGivenParentDataset method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.getParentFilesGivenParentDataset(PARENT_DATASET,
                                                            [FILE1_WITH_PARENT, FILE2_WITH_PARENT])
        self.assertItemsEqual(results[FILE1_WITH_PARENT], {'/store/data/ComissioningHI/Cosmics/RAW/v1/000/180/841/721E482F-A407-E111-8C0C-BCAEC518FF6E.root'})
        self.assertItemsEqual(results[FILE2_WITH_PARENT], {'/store/data/ComissioningHI/Cosmics/RAW/v1/000/180/852/C0A3D337-1408-E111-836F-0030486780A8.root'})

    def testGetParentFilesByLumi(self):
        """Test the getParentFilesByLumi method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.getParentFilesByLumi(FILE1_WITH_PARENT)
        self.assertItemsEqual(results[0]['ParentDataset'], PARENT_DATASET)
        self.assertItemsEqual(results[0]['ParentFiles'], ['/store/data/ComissioningHI/Cosmics/RAW/v1/000/180/841/721E482F-A407-E111-8C0C-BCAEC518FF6E.root'])

    @attr('integration')  # too much data to mock
    def testListBlocksWithNoParents(self):
        """Test the listBlocksWithNoParents method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.listBlocksWithNoParents(DATASET_WITH_PARENTS)
        self.assertItemsEqual(results, [])

        results = self.dbs.listBlocksWithNoParents(DATASET)
        self.assertEqual(len(results), 46)
        self.assertTrue('/HighPileUp/Run2011A-v1/RAW#6ffd4f16-cc42-11e0-80a9-003048caaace' in results)

    def testListFilesWithNoParents(self):
        """Test the listFilesWithNoParents method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.listFilesWithNoParents(BLOCK_WITH_PARENTS)
        self.assertEqual(len(results), 0)

        results = self.dbs.listFilesWithNoParents(BLOCK)
        self.assertEqual(len(results), 2)
        self.assertTrue('/store/data/Run2011A/HighPileUp/RAW/v1/000/173/657/B293AF24-BFCB-E011-8F85-BCAEC5329701.root' in results)

    def testGetParentDatasetTrio(self):
        """Test the getParentDatasetTrio method"""
        self.dbs = DBSReader(self.endpoint)
        results = self.dbs.getParentDatasetTrio(DATASET_WITH_PARENTS)
        self.assertTrue((180851, 2) in results)
        self.assertEqual(int(results[(180851, 2)]), 36092526)
        self.assertTrue((180851, 1) in results)
        self.assertEqual(int(results[(180851, 1)]), 36092526)

    def testCompileParentageList(self):
        """Test the _compileParentageList method"""
        self.dbs = DBSReader(self.endpoint)
        # it contains 3 files
        childDset = "/HToAATo4L_H5000A10_TuneCP5_13TeV-pythia8/RunIISummer20UL16MiniAODAPVv2-106X_mcRun2_asymptotic_preVFP_v11-v2/MINIAODSIM"
        childBlock = "/HToAATo4L_H5000A10_TuneCP5_13TeV-pythia8/RunIISummer20UL16MiniAODAPVv2-106X_mcRun2_asymptotic_preVFP_v11-v2/MINIAODSIM#bf137a20-65c5-4927-8f62-7733c3d9e017"
        parentRunLumi = self.dbs.getParentDatasetTrio(childDset)

        expectRes = [['4013778997', '4005114237'], ['4013778997', '4005114277'], ['4013778997', '4005114357'],
                     ['4013778997', '4005114397'], ['4013779037', '4005114157'], ['4013779037', '4005114197'],
                     ['4013779037', '4005114317'], ['4013779037', '4005114357'], ['4013779077', '4005114437'],
                     ['4013779077', '4005114477'], ['4013779077', '4005114517']]
        childParent, missingFiles = self.dbs._compileParentageList(childBlock, parentRunLumi)
        self.assertItemsEqual(expectRes, childParent)



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_SetupCMSSWPset_t.py

Tests for the PSet configuration code.

"""


from builtins import zip

import unittest
import os
import sys
import shutil

from WMCore.DataStructs.File import File
from WMCore.DataStructs.Job import Job
from WMCore.Configuration import ConfigSection
from WMCore.Storage.TrivialFileCatalog import loadTFC
from WMCore.WMSpec.WMStep import WMStep
from WMCore.WMSpec.Steps import StepFactory
from WMCore.WMSpec.Steps.Fetchers.PileupFetcher import PileupFetcher
from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig
from WMQuality.TestInit import TestInit
import WMCore.WMBase


class SetupCMSSWPsetTest(unittest.TestCase):
    def setUp(self):
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testDir = self.testInit.generateWorkDir()
        sys.path.insert(0, os.path.join(WMCore.WMBase.getTestBase(),
                                        "WMCore_t/WMRuntime_t/Scripts_t"))

    def tearDown(self):
        sys.path.remove(os.path.join(WMCore.WMBase.getTestBase(), "WMCore_t/WMRuntime_t/Scripts_t"))
        if 'WMTaskSpace' in sys.modules:
            del sys.modules["WMTaskSpace"]
        self.testInit.delWorkDir()
        os.unsetenv("WMAGENT_SITE_CONFIG_OVERRIDE")


    def createTestStep(self):
        """
        _createTestStep_

        Create a test step that can be passed to the setup script.

        """
        newStep = WMStep("cmsRun1")
        stepTemplate = StepFactory.getStepTemplate("CMSSW")
        stepTemplate.install(newStep)
        newStepHelper = stepTemplate.helper(newStep)

        newStepHelper.setStepType("CMSSW")
        newStepHelper.setGlobalTag("SomeGlobalTag")
        newStepHelper.data.application.section_("setup")
        newStepHelper.cmsswSetup("CMSSW_11_0_2", scramArch=['slc7_amd64_gcc820'])

        return newStepHelper


    def createTestJob(self):
        """
        _createTestJob_

        Create a test job that has parents for each input file.

        """
        newJob = Job(name = "TestJob")
        newJob.addFile(File(lfn = "/some/file/one",
                            parents = set([File(lfn = "/some/parent/one")])))
        newJob.addFile(File(lfn = "/some/file/two",
                            parents = set([File(lfn = "/some/parent/two")])))
        return newJob

    def loadProcessFromPSet(self, psetPath=None):
        """
        _loadProcessFromPSet_

        This requires changing the working directory,
        do so in a safe manner to encapsulate the change to this method only
        """
        from WMCore.WMRuntime.Scripts.SetupCMSSWPset import Unpickler
        currentPath = os.getcwd()
        loadedProcess = None

        if psetPath is None:
            psetPath = self.testDir
        with open(os.path.join(psetPath, "PSet.pkl"), "rb") as f:
            pset = Unpickler(f).load()

        os.chdir(currentPath)

        return pset

    def testPSetFixup(self):
        """
        _testPSetFixup_

        Verify that all necessary parameters are set in the PSet.

        """
        from WMCore.WMRuntime.Scripts.SetupCMSSWPset import SetupCMSSWPset

        setupScript = SetupCMSSWPset()
        setupScript.step = self.createTestStep()
        setupScript.stepSpace = ConfigSection(name = "stepSpace")
        setupScript.stepSpace.location = self.testDir
        shutil.copyfile(os.path.join(os.path.dirname(__file__), "WMTaskSpace", "cmsRun1", "PSet.py"),
                        os.path.join(setupScript.stepSpace.location, "PSet.py"))
        setupScript.job = self.createTestJob()
        setupScript()

        fixedPSet = self.loadProcessFromPSet(setupScript.stepSpace.location)

        self.assertTrue(hasattr(fixedPSet.source, 'fileNames'))
        self.assertTrue(hasattr(fixedPSet.source, 'secondaryFileNames'))
        self.assertEqual(fixedPSet.maxEvents.input._value, -1,
                         "Error: Wrong maxEvents.")

    def testEventsPerLumi(self):
        """
        _testEventsPerLumi_
        Verify that you can put in events per lumi in the process.

        """
        from WMCore.WMRuntime.Scripts.SetupCMSSWPset import SetupCMSSWPset

        setupScript = SetupCMSSWPset()
        setupScript.step = self.createTestStep()
        setupScript.step.setEventsPerLumi(500)
        setupScript.stepSpace = ConfigSection(name = "stepSpace")
        setupScript.stepSpace.location = self.testDir
        shutil.copyfile(os.path.join(os.path.dirname(__file__), "WMTaskSpace", "cmsRun1", "PSet.py"),
                        os.path.join(setupScript.stepSpace.location, "PSet.py"))
        setupScript.job = self.createTestJob()
        setupScript()

        fixedPSet = self.loadProcessFromPSet(setupScript.stepSpace.location)

        self.assertTrue(hasattr(fixedPSet.source, 'fileNames'))
        self.assertTrue(hasattr(fixedPSet.source, 'secondaryFileNames'))
        self.assertEqual(fixedPSet.source.numberEventsInLuminosityBlock._value,
                         500, "Error: Wrong number of events per luminosity block")
        self.assertEqual(fixedPSet.maxEvents.input._value, -1,
                         "Error: Wrong maxEvents.")

    def testChainedProcesing(self):
        """
        test for chained CMSSW processing - check the overriden TFC, its values
        and that input files is / are set correctly.

        """
        from WMCore.WMRuntime.Scripts.SetupCMSSWPset import SetupCMSSWPset

        setupScript = SetupCMSSWPset()
        setupScript.step = self.createTestStep()
        setupScript.stepSpace = ConfigSection(name = "stepSpace")
        setupScript.stepSpace.location = self.testDir
        shutil.copyfile(os.path.join(os.path.dirname(__file__), "WMTaskSpace", "cmsRun1", "PSet.py"),
                        os.path.join(setupScript.stepSpace.location, "PSet.py"))
        setupScript.job = self.createTestJob()
        setupScript.step.setupChainedProcessing("my_first_step", "my_input_module")
        setupScript()
        fixedPSet = self.loadProcessFromPSet(setupScript.stepSpace.location)

        # test if the overriden TFC is right
        self.assertFalse(hasattr(setupScript.step.data.application, "overrideCatalog"),
                        "We no longer override the TFC, instead only make a PSet tweak!")


    def testPileupSetup(self):
        """
        Test the pileup setting.

        reference (setupScript.process instance):
        in test/python/WMCore_t/WMRuntime_t/Scripts_t/WMTaskSpace/cmsRun1/PSet.py

        """
        try:
            from dbs.apis.dbsClient import DbsApi
        except ImportError as ex:
            raise unittest.SkipTest

        # this is modified and shortened version of
        # WMCore/test/python/WMCore_t/Misc_t/site-local-config.xml
        # since the dataset name in question (below) is only present at
        # storm-fe-cms.cr.cnaf.infn.it, need to make the test think it's its local SE
        siteLocalConfigContent = \
        """
<site-local-config>
    <site name="-SOME-SITE-NAME-">
        <event-data>
            <catalog url="trivialcatalog_file:/uscmst1/prod/sw/cms/SITECONF/T1_US_FNAL/PhEDEx/storage.xml?protocol=dcap"/>
        </event-data>
        <local-stage-out>
            <!-- original cmssrm.fnal.gov -->
            <phedex-node value="T2_CH_CERN"/>
            <command value="test-copy"/>
            <catalog url="trivialcatalog_file:/uscmst1/prod/sw/cms/SITECONF/T1_US_FNAL/PhEDEx/storage.xml?protocol=dcap"/>
        </local-stage-out>
        <calib-data>
            <frontier-connect>
                <load balance="proxies"/>
                <proxy url="http://cmsfrontier1.fnal.gov:3128"/>
                <proxy url="http://cmsfrontier2.fnal.gov:3128"/>
            </frontier-connect>
        </calib-data>
    </site>
</site-local-config>
"""
        siteLocalConfig = os.path.join(self.testDir, "test-site-local-config.xml")
        f = open(siteLocalConfig, 'w')
        f.write(siteLocalConfigContent)
        f.close()

        from WMCore.WMRuntime.Scripts.SetupCMSSWPset import SetupCMSSWPset
        setupScript = SetupCMSSWPset()
        setupScript.step = self.createTestStep()
        setupScript.stepSpace = ConfigSection(name = "stepSpace")
        setupScript.stepSpace.location = os.path.join(self.testDir, "cmsRun1")
        setupScript.job = self.createTestJob()
        # define pileup configuration
        # despite of the implementation considering whichever type of pileup,
        # only "data" and "mc" types are eventually considered and lead to any
        # modifications of job input files
        pileupConfig = {"data": ["/Mu/PenguinsPenguinsEverywhere-SingleMu-HorriblyJaundicedYellowEyedPenginsSearchingForCarrots-v31/RECO"],
                        "mc": ["/Mu/PenguinsPenguinsEverywhere-SingleMu-HorriblyJaundicedYellowEyedPenginsSearchingForCarrots-v31/RECO"]}
        dbsUrl = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"
        setupScript.step.setupPileup(pileupConfig, dbsUrl)
        # SetupCMSSWPset pileup handling will be consulting SiteLocalConfig
        # to determine StorageElement (SE) name the job is running on
        # SiteLocalConfig loads the site-local-config.xml file from env.
        # variable defined location ; if the variable is not defined already, set it
        # obviously, if "WMAGENT_SITE_CONFIG_OVERRIDE" is already set here, the above
        # thick with SE name is not effective
        if not os.getenv("WMAGENT_SITE_CONFIG_OVERRIDE", None):
            os.environ["WMAGENT_SITE_CONFIG_OVERRIDE"] = siteLocalConfig
        # find out local site name from the testing local site config,
        # will be needed later
        siteConfig = loadSiteLocalConfig()
        seLocalName = siteConfig.localStageOut["phedex-node"]
        print("Running on site '%s', local SE name: '%s'" % (siteConfig.siteName, seLocalName))

        # before calling the script, SetupCMSSWPset will try to load JSON
        # pileup configuration file, need to create it in self.testDir
        fetcher = PileupFetcher()
        fetcher.setWorkingDirectory(self.testDir)
        fetcher.createPileupConfigFile(setupScript.step)

        setupScript()

        # now test all modifications carried out in SetupCMSSWPset.__call__
        # which will also test that CMSSWStepHelper.setupPileup run correctly
        mixModules, dataMixModules = setupScript._getPileupMixingModules()

        # load in the pileup configuration in the form of dict which
        # PileupFetcher previously saved in a JSON file
        pileupDict = setupScript._getPileupConfigFromJson()

        # get the sub dict for particular pileup type
        # for pileupDict structure description - see PileupFetcher._queryDbsAndGetPileupConfig
        for pileupType, modules in zip(("data", "mc"), (dataMixModules, mixModules)):
            # getting KeyError here - above pileupConfig is not correct - need
            # to have these two types of pile type
            d = pileupDict[pileupType]
            self._mixingModulesInputFilesTest(modules, d, seLocalName)


    def _mixingModulesInputFilesTest(self, modules, pileupSubDict, seLocalName):
        """
        pileupSubDic - contains only dictionary for particular pile up type

        """
        # consider only locally available files
        filesInConfigDict = []
        for v in pileupSubDict.values():
            if seLocalName in v["phedexNodeNames"]:
                filesInConfigDict.extend(v["FileList"])

        for m in modules:
            inputTypeAttrib = getattr(m, "input", None) or getattr(m, "secsource", None)
            fileNames = inputTypeAttrib.fileNames.value
            if fileNames == None:
                fileNames = []
            m = ("Pileup configuration file list '%s' and mixing modules input "
                 "filelist '%s' are not identical." % (filesInConfigDict, fileNames))
            self.assertEqual(sorted(filesInConfigDict), sorted(fileNames), m)



if __name__ == "__main__":
    unittest.main()



# Auto generated configuration file
# using:
# Revision: 1.392
# Source: /local/reps/CMSSW/CMSSW/Configuration/PyReleaseValidation/python/ConfigBuilder.py,v
# with command line options: step3 --datatier GEN-SIM-RECO,DQM --conditions auto:startup -s RAW2DIGI,L1Reco,RECO,VALIDATION,DQM --eventcontent RECOSIM,DQM --io RECO.io --python RECO.py -n 100 --no_exec --filein file:step2.root --fileout file:step3.root
# modified to reduce outside dependencies - this probably doesn't do anything useful anymore
# hopefully its still useful as a test

import FWCore.ParameterSet.Config as cms

process = cms.Process('RECO')

# import of standard configurations
#process.load('Configuration.StandardSequences.Services_cff')
#process.load('SimGeneral.HepPDTESSource.pythiapdt_cfi')
#process.load('FWCore.MessageService.MessageLogger_cfi')
#process.load('Configuration.EventContent.EventContent_cff')
#process.load('SimGeneral.MixingModule.mixNoPU_cfi')
#process.load('Configuration.StandardSequences.GeometryRecoDB_cff')
#process.load('Configuration.StandardSequences.MagneticField_38T_cff')
#process.load('Configuration.StandardSequences.RawToDigi_cff')
#process.load('Configuration.StandardSequences.L1Reco_cff')
#process.load('Configuration.StandardSequences.Reconstruction_cff')
#process.load('Configuration.StandardSequences.Validation_cff')
#process.load('DQMOffline.Configuration.DQMOfflineMC_cff')
#process.load('Configuration.StandardSequences.EndOfProcess_cff')
#process.load('Configuration.StandardSequences.FrontierConditions_GlobalTag_cff')

process.maxEvents = cms.untracked.PSet(
    input = cms.untracked.int32(100)
)

# Input source
process.source = cms.Source("PoolSource",
    secondaryFileNames = cms.untracked.vstring(),
    fileNames = cms.untracked.vstring('file:step2.root')
)

process.options = cms.untracked.PSet(

)

# Production Info
process.configurationMetadata = cms.untracked.PSet(
    version = cms.untracked.string('$Revision: 1.392 $'),
    annotation = cms.untracked.string('step3 nevts:100'),
    name = cms.untracked.string('PyReleaseValidation')
)

# Output definition

process.RECOSIMoutput = cms.OutputModule("PoolOutputModule",
    splitLevel = cms.untracked.int32(0),
    eventAutoFlushCompressedSize = cms.untracked.int32(5242880),
#    outputCommands = process.RECOSIMEventContent.outputCommands,
    fileName = cms.untracked.string('file:step3.root'),
    dataset = cms.untracked.PSet(
        filterName = cms.untracked.string(''),
        dataTier = cms.untracked.string('GEN-SIM-RECO')
    )
)

process.DQMoutput = cms.OutputModule("PoolOutputModule",
    splitLevel = cms.untracked.int32(0),
#    outputCommands = process.DQMEventContent.outputCommands,
    fileName = cms.untracked.string('file:step3_inDQM.root'),
    dataset = cms.untracked.PSet(
        filterName = cms.untracked.string(''),
        dataTier = cms.untracked.string('DQM')
    )
)

# Additional output definition

# Other statements
#process.mix.playback = True
#process.mix.digitizers = cms.PSet()
for a in process.aliases: delattr(process, a)
#process.RandomNumberGeneratorService.restoreStateLabel=cms.untracked.string("randomEngineStateProducer")
#process.GlobalTag.globaltag = 'START60_V4::All'

# Path and EndPath definitions
process.raw2digi_step = cms.Path() #cms.Path(process.RawToDigi)
process.L1Reco_step = cms.Path() #cms.Path(process.L1Reco)
process.reconstruction_step = cms.Path() #cms.Path(process.reconstruction)
process.prevalidation_step = cms.Path() #cms.Path(process.prevalidation)
process.dqmoffline_step = cms.Path() #cms.Path(process.DQMOffline)
process.validation_step = cms.EndPath() #cms.EndPath(process.validation)
process.endjob_step = cms.EndPath() #cms.EndPath(process.endOfProcess)
process.RECOSIMoutput_step = cms.EndPath() #cms.EndPath(process.RECOSIMoutput)
process.DQMoutput_step = cms.EndPath() #cms.EndPath(process.DQMoutput)

# Schedule definition
process.schedule = cms.Schedule(process.raw2digi_step,process.L1Reco_step,process.reconstruction_step,process.prevalidation_step,process.validation_step,process.dqmoffline_step,process.endjob_step,process.RECOSIMoutput_step,process.DQMoutput_step)



"""
Created on Jun 18, 2009

@author: meloam

Modified on Nov. 7, 2023 by Duong Nguyen
"""
import os
import unittest

from WMCore.Storage.DeleteMgr import DeleteMgr
# need this otherwise retrieveStageOutImpl fails
from WMCore.Storage.Registry import retrieveStageOutImpl
from WMCore.Storage.StageOutError import StageOutFailure


class DeleteMgrTest(DeleteMgr):
    def deletePFN(self, pfn, lfn, command):
        """
        Delete the given PFN
        """
        try:
            impl = retrieveStageOutImpl(command)
        except Exception as ex:
            msg = "Unable to retrieve impl for file deletion in %s\n" % (pfn)
            msg += "Error retrieving StageOutImpl for command named: %s\n" % (
                command,)
            raise StageOutFailure(msg, Command=command,
                                  LFN=lfn, ExceptionDetail=str(ex))
        impl.numRetries = self.numberOfRetries
        impl.retryPause = self.retryPauseTime

        try:
            pass
        except Exception as ex:
            self.logger.exception("Failed to delete file: %s", pfn)
            raise ex

        return pfn


class DeleteMgrUnitTest(unittest.TestCase):

    def setUp(self):
        # shut up SiteLocalConfig
        os.putenv('CMS_PATH', os.getcwd())
        os.putenv('SITECONFIG_PATH', os.getcwd())

    def testDeleteMgr(self):
        os.environ['SITECONFIG_PATH'] = '/cvmfs/cms.cern.ch/SITECONF/T1_US_FNAL'
        deleteMgr = DeleteMgrTest()
        # keep using 'phedex-node' for overrideParams (to be compatible with the whole DMWM structure?)
        deleteMgr_override = DeleteMgrTest(
            **{"command": "gfal2", "phedex-node": "T1_US_FNAL_Disk", "lfn-prefix": "root://abc/xyz"})
        fileToDelete = {'LFN': '/store/abc/xyz.root'}
        deleteMgr(fileToDelete)
        deleteMgr_override(fileToDelete)

        return


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#!/usr/bin/env python
"""
_WorkerThreads_

Module to handle regular worker threads. To use, create an instance of
WorkerThreadManager to manage the worker threads. Derive worker classes from
BaseWorkerThread, and override at least the algorithm(self, parameters) method.
Run the workers by calling the WorkerThreadManager.addWorker method. Worker
threads have access to the parent thread dbFactory and component as standard.
The usual default transaction object (and transaction dictionary), trigger and
message service are attached to the child thread by the underlying framework.
E.g.:

from WMCore.WorkerThreads.WorkerThreadManager import WorkerThreadManager
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
import time, threading

# Create a new worker object
class MyWorker(BaseWorkerThread):
    def __init__(self):
        BaseWorkerThread.__init__(self)

    def algorithm(self, parameters):
        print "Doing work with parameters:", parameters
        myThread = threading.currentThread()
        myThread.transaction.begin()
        # Do DB work here
        myThread.transaction.commit()

    def setup(self, parameters):
        print "Setting up worker with parameters:", parameters

    def terminate(self, parameters):
        print "Terminating worker with parameters:", parameters

# Create and pause a worker manager
manager = WorkerThreadManager(parentComponent)
manager.pauseWorkers()

# Add a MyWorker() instance to run every 2 seconds and one every 3 seconds
manager.addWorker(MyWorker(), 2, ["Some", "Parameters"])
manager.addWorker(MyWorker(), 3, ["Some again", "Parameters again"])

# Run the worker threads
manager.resumeWorkers()
time.sleep(5)

# Terminate all worker threads
manager.terminateWorkers()

"""



#!/usr/bin/env python
"""
_RucioConMon_t_

Unit tests for RucioConMon WMCore Service class
"""

import unittest

from nose.plugins.attrib import attr

from WMCore.Services.RucioConMon.RucioConMon import RucioConMon


class RucioConMonTest(unittest.TestCase):
    """
    Unit tests for RucioConMon Service module
    """

    @attr("integration")
    def testGetRSEUnmerged(self):
        """
        Test getRSEUnmerged method using both zipped and unzipped requests
        This test uses specific rse name which can be changed to any other RSE.
        """
#         url = "https://cmsweb.cern.ch/rucioconmon/WM/files?rse=T2_TR_METU&format=raw"
        mgr = RucioConMon("https://cmsweb.cern.ch/rucioconmon")
        rseName = "T2_TR_METU"
        dataUnzipped = mgr.getRSEUnmerged(rseName, zipped=False)
        dataZipped = mgr.getRSEUnmerged(rseName, zipped=True)
        self.assertTrue(dataUnzipped == dataZipped)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
_XMLParser_t_

Unit tests for XMLParser module
"""

# system modules
import os
import unittest

# WMCore modules
from WMCore.FwkJobReport.XMLParser import perfSummaryHandler, perfRepHandler, \
        reportBuilder, reportDispatcher, inputFileHandler, fileHandler, \
        runHandler, branchHandler, inputAssocHandler, \
        perfCPUHandler, perfMemHandler, perfStoreHandler, castMetricValue
from WMCore.FwkJobReport.Report import Report
from WMCore.Algorithms.ParseXMLFile import xmlFileToNode
from WMCore.WMBase import getTestBase


class XMLParserTest(unittest.TestCase):
    """
    _XMLParserTest_

    Unit tests for XMLParser module
    """

    def setUp(self):
        """
        _setUp_

        Figure out the location of the XML report produced by CMSSW.
        """
        testData = os.path.join(getTestBase(), "WMCore_t/FwkJobReport_t")
        self.xmlFile = os.path.join(testData, "CMSSWMergeReport2.xml")
        self.xmlXrd = os.path.join(testData, "CMSSWJobReportXrdSiteStatistics.xml")

    def testPerfSummaryHandler(self):
        """
        testPerfSummaryHandler
        Check performance summary handler function
        """
        # read XML, build node structure
        node = xmlFileToNode(self.xmlFile)

        # Set up coroutine pipeline
        fileDispatchers = {
            "Runs": runHandler(),
            "Branches": branchHandler(),
            "Inputs": inputAssocHandler(),
        }
        perfRepDispatchers = {
            "CPU": perfCPUHandler(),
            "Memory": perfMemHandler(),
            "Storage": perfStoreHandler(),
            "PerformanceSummary": perfSummaryHandler()
        }
        dispatchers = {
            "File": fileHandler(fileDispatchers),
            "InputFile": inputFileHandler(fileDispatchers),
            "PerformanceReport": perfRepHandler(perfRepDispatchers)
        }

        # Feed pipeline with node structure and report result instance
        report = Report('cmsRun1')
        report.parse(self.xmlFile)
        reportBuilder(node, report, reportDispatcher(dispatchers))
        cmsRun = getattr(report.data, 'cmsRun1', {})
        performance = getattr(cmsRun, 'performance', {})
        cmssw = getattr(performance, 'cmssw', {})
        obj = cmssw.dictionary_()
        keys = ['SystemMemory', 'ProcessingSummary', 'StorageStatistics', 'SystemCPU', 'Timing', 'ApplicationMemory']
        for key in obj.keys():
            self.assertTrue(key in keys)

    def testXrdSiteStatistics(self):
        """
        unit test for XrdSiteStatistics, it covers castXrdSiteStatistics function.
        Check XrdSiteStatistics metrics in performance summary part of FJR report
        """
        # Feed pipeline with node structure and report result instance
        report = Report('cmsRun1')
        report.parse(self.xmlXrd)
        cmsRun = getattr(report.data, 'cmsRun1', {})
        performance = getattr(cmsRun, 'performance', {})
        cmssw = getattr(performance, 'cmssw', {})
        xrd = getattr(cmssw, 'XrdSiteStatistics', {})
        rdict = xrd.dictionary_()
        keys = [
                'read-numOperations',
                'read-totalMegabytes',
                'readv-numChunks',
                'readv-numOperations',
                'readv-totalMegabytes',
                'readv-totalMsecs'
                ]
        for key in keys:
            self.assertTrue(key in rdict.keys())
            for idict in rdict.get(key, []):
                self.assertTrue('site' in idict.keys())
                self.assertTrue('value' in idict.keys())

    def testCastMetricValue(self):
        """
        unit test castMetricValue function
        """
        self.assertTrue(1, castMetricValue("1"))
        self.assertTrue(1.1, castMetricValue("1.1"))
        self.assertTrue(True, castMetricValue("true"))
        self.assertTrue("bla", castMetricValue("bla "))



if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
AgentStatusWatcher test
"""

from __future__ import print_function, division

import threading
import unittest
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit

from WMComponent.AgentStatusWatcher.AgentStatusPoller import AgentStatusPoller


class AgentStatusWatcher_t(unittest.TestCase):
    """
    TestCase for AgentStatustWatcher module
    """

    def setUp(self):
        """
        setup for test.
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.reqmonDBName = "wmstat_t"
        self.localDBName = "wmstat_t_local"
        self.testInit.setupCouch(self.reqmonDBName, "WMStatsAgent")
        self.testInit.setupCouch(self.localDBName, "WMStatsAgent")
        self.testDir = self.testInit.generateWorkDir()
        return

    def tearDown(self):
        """
        Database deletion
        """
        threading.currentThread()

        self.testInit.delWorkDir()
        self.testInit.tearDownCouch()
        self.testInit.clearDatabase()
        return

    def getConfig(self):
        """
        _createConfig_

        General config file
        """

        couchURL = self.testInit.couchUrl
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        config.section_("Agent")
        config.Agent.hostName = "localhost"
        config.Agent.contact = "test@test.com"
        config.Agent.teamName = "testTeam"
        config.Agent.agentName = "testAgentName"
        config.Agent.useMsgService = False
        config.Agent.useTrigger = False
        config.Agent.useHeartbeat = False

        config.component_("AgentStatusWatcher")
        config.AgentStatusWatcher.namespace = "WMComponent.AgentStatusWatcher.AgentStatusWatcher"
        config.AgentStatusWatcher.jsonFile = "/dev/null"

        config.component_("AnalyticsDataCollector")
        config.AnalyticsDataCollector.namespace = "WMComponent.AnalyticsDataCollector.AnalyticsDataCollector"
        config.AnalyticsDataCollector.componentDir = config.General.workDir + "/AnalyticsDataCollector"
        config.AnalyticsDataCollector.localQueueURL = "%s/%s" % (couchURL, "workqueue")
        config.AnalyticsDataCollector.summaryLevel = "task"

        return config

    def testAgentStatusPoller(self):
        """
        Tests the components, as in sees if they load.
        Otherwise does nothing.
        """
        threading.currentThread()
        config = self.getConfig()
        watcher = AgentStatusPoller(config)
        # Note:
        # Should call: watcher.setup() and watcher.algorithm()
        # for a complete poller test.
        # Just testing proxy/certificate lifetimes for now.

        # Start with ok status
        agInfo = {'status':'ok'}

        # Check service certificate and proxy lifetimes
        watcher.checkCredLifetime(agInfo, "proxy")
        watcher.checkCredLifetime(agInfo, "certificate")
        print("Agent Info:\n%s" % (agInfo))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
InstanceTestPage.py

"""
from cherrypy import expose
from WMCore.WebTools.Page import Page

class InstanceTestPage(Page):
    @expose
    def index(self):
        return self.config.instance

    @expose
    def default(self, *args, **kwargs):
        return self.index()

    @expose
    def database(self):
        return self.config.database.connectUrl

    @expose
    def security(self):
        return self.config.security.sec_params



"""
Created on May 19, 2015
"""

from WMCore.REST.CherryPyPeriodicTask import CherryPyPeriodicTask
from WMCore.ReqMgr.DataStructs.RequestStatus import AUTO_TRANSITION, CANCEL_AUTO_TRANSITION
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue
from WMCore.Services.ReqMgr.ReqMgr import ReqMgr


def moveForwardStatus(reqmgrSvc, wfStatusDict, logger):

    for status, nextStatus in AUTO_TRANSITION.items():
        count = 0
        requests = reqmgrSvc.getRequestByStatus([status], detail=False)
        for wf in requests:
            stateFromGQ = wfStatusDict.get(wf, None)
            if stateFromGQ in [None, "canceled"]:
                continue
            elif stateFromGQ == status:
                continue
            elif stateFromGQ == "failed" and status == "staged":
                count += 1
                reqmgrSvc.updateRequestStatus(wf, nextStatus[0])
                logger.info("%s in %s moved to %s", wf, status, nextStatus[0])
                continue
            elif stateFromGQ == "failed" and status == "acquired":
                count += 1
                reqmgrSvc.updateRequestStatus(wf, stateFromGQ)
                logger.info("%s in %s moved to %s", wf, status, stateFromGQ)
                continue

            try:
                i = nextStatus.index(stateFromGQ)
            except ValueError:
                # No state change needed
                continue
            for j in range(i + 1):
                count += 1
                reqmgrSvc.updateRequestStatus(wf, nextStatus[j])
                logger.info("%s in %s moved to %s", wf, status, nextStatus[j])
        logger.info("%s requests moved to new state from %s", count, status)
    return


def moveToCompletedForNoWQJobs(reqmgrSvc, globalQSvc, wfStatusDict, logger):
    """
    Handle workflows that have been either aborted or force-completed.
    This will ensure that no global workqueue elements will be left behind.

    :param reqmgrSvc: object instance of the ReqMgr class
    :param globalQSvc: object instance of the WorkQueue class
    :param wfStatusDict: workflow status according to the workqueue elements
    :param logger: a logger object instance
    :return: None object
    """
    for status, nextStatus in CANCEL_AUTO_TRANSITION.items():
        requests = reqmgrSvc.getRequestByStatus([status], detail=False)
        count = 0
        for wflowName in requests:
            stateFromGQ = wfStatusDict.get(wflowName, None)
            if stateFromGQ == "canceled":
                # elements still in CancelRequested, wait for the agent to do his job
                continue
            elif stateFromGQ in ["acquired", "running-open", "running-closed"]:
                # then something went wrong with the workflow abortion/force-completion
                # trigger another cancel request
                logger.info("%s in %s but WQEs in %s, cancelling it again!",
                            wflowName, status, stateFromGQ)
                globalQSvc.cancelWorkflow(wflowName)
            elif stateFromGQ in ["completed", None]:
                # all elements are already in a final state or no longer exist, advance status
                count += 1
                reqmgrSvc.updateRequestStatus(wflowName, nextStatus)
                logger.info("%s in %s moved to %s", wflowName, status, nextStatus)
        logger.info("Total %s: %d", nextStatus, count)


def moveOpenRequests(globalQSvc, wfStatusDict, logger):
    """
    Method to advance the request status for workflows that must remain
    open for new work in global workqueue.
    :param globalQSvc: object instance of the WorkQueue class
    :param wfStatusDict: workflow status according to the workqueue elements
    :param logger: a logger object instance
    :return: None object, but the wfStatusDict object can be updated in this function.
    """
    logger.info("Getting requests open for new data")
    openWflows = globalQSvc.getOpenRequests(inboxFlag=True)
    logger.info("Found %d requests still opened in global workqueue.", len(openWflows))
    for wflow in openWflows:
        if wflow not in wfStatusDict:
            msg = f"Workflow {wflow} marked as open for new data, but it"
            msg += "has no status based on the set of WQEs. Skipping it."
            logger.info(msg)
        elif wfStatusDict.get(wflow, "") in ("running-closed", "completed"):
            logger.info(f"Workflow {wflow} being forced to stay in running-open")
            # open workflows should not advance beyond running-open
            wfStatusDict[wflow] = "running-open"
        # otherwise let it follow its natural status transition
    return


class StatusChangeTasks(CherryPyPeriodicTask):
    def __init__(self, rest, config):
        super(StatusChangeTasks, self).__init__(config)

    def setConcurrentTasks(self, config):
        """
        Define which function to periodically run
        """
        self.concurrentTasks = [{'func': self.advanceStatus, 'duration': config.checkStatusDuration}]

    def advanceStatus(self, config):
        """
        Advance the request status based on the global workqueue elements status
        """
        reqmgrSvc = ReqMgr(config.reqmgr2_url, logger=self.logger)
        globalQSvc = WorkQueue(config.workqueue_url)

        self.logger.info("Getting GQ data for status check")
        wfStatusDict = globalQSvc.getWorkflowStatusFromWQE()
        moveOpenRequests(globalQSvc, wfStatusDict, self.logger)

        self.logger.info("Advancing statuses")
        moveForwardStatus(reqmgrSvc, wfStatusDict, self.logger)
        moveToCompletedForNoWQJobs(reqmgrSvc, globalQSvc, wfStatusDict, self.logger)

        self.logger.info("Done advancing status")

        return
